from secrets import token_bytes

import pytest

import pyseto
from pyseto import DecryptError, EncryptError, Key, SignError, VerifyError
from pyseto.versions.v3 import V3Local, V3Public

from .utils import load_key


class TestV3Local:
    """
    Tests for v3.local.
    """

    @pytest.mark.parametrize(
        "key, msg",
        [
            (b"", "key must be specified."),
        ],
    )
    def test_v3_local_new_with_invalid_arg(self, key, msg):
        with pytest.raises(ValueError) as err:
            Key.new(3, "local", key)
            pytest.fail("Key.new() should fail.")
        assert msg in str(err.value)

    def test_v3_local_decrypt_via_decode_with_wrong_key(self):
        k1 = Key.new(3, "local", b"our-secret")
        k2 = Key.new(3, "local", b"others-secret")
        token = pyseto.encode(k1, b"Hello world!")
        with pytest.raises(DecryptError) as err:
            pyseto.decode(k2, token)
            pytest.fail("pyseto.decode() should fail.")
        assert "Failed to decrypt." in str(err.value)

    def test_v3_local_encrypt_with_invalid_arg(self):
        k = Key.new(3, "local", b"our-secret")
        with pytest.raises(EncryptError) as err:
            k.encrypt(None)
            pytest.fail("pyseto.encrypt() should fail.")
        assert "Failed to encrypt." in str(err.value)

    @pytest.mark.parametrize(
        "nonce",
        [
            token_bytes(1),
            token_bytes(8),
            token_bytes(31),
            token_bytes(33),
            token_bytes(64),
        ],
    )
    def test_v3_local_encrypt_via_encode_with_wrong_nonce(self, nonce):
        k = Key.new(3, "local", b"our-secret")
        with pytest.raises(ValueError) as err:
            pyseto.encode(k, b"Hello world!", nonce=nonce)
            pytest.fail("pyseto.encode() should fail.")
        assert "nonce must be 32 bytes long." in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.local.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k4.local.AAAAAAAAAAAAAAAA", "Invalid PASERK version: k4."),
            ("k3.local.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k3.public.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k3.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
            ("k3.public.AAAAAAAAAAAAAAAA", "Invalid PASERK type: public."),
            (
                "k3.local-wrap.AAAAAAAAAAAAAAAA",
                "local-wrap needs wrapping_key.",
            ),
            (
                "k3.secret-wrap.AAAAAAAAAAAAAAAA",
                "Invalid PASERK type: secret-wrap.",
            ),
            (
                "k3.local-pw.AAAAAAAAAAAAAAAA",
                "local-pw needs password.",
            ),
        ],
    )
    def test_v3_local_from_paserk_with_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V3Local.from_paserk(paserk)
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.local-wrap.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k3.local-wrap.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k3.local-wrap.xxx.AAAAAAAAAAAAAAAA", "Unknown wrapping algorithm: xxx."),
            ("k3.xxx.pie.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
        ],
    )
    def test_v3_local_from_paserk_with_wrapping_key_and_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V3Local.from_paserk(paserk, wrapping_key=token_bytes(32))
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    # @pytest.mark.parametrize(
    #     "paserk, msg",
    #     [
    #         ("k3.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
    #     ],
    # )
    # def test_v3_local_from_paserk_with_unsealing_key_and_invalid_args(
    #     self, paserk, msg
    # ):

    #     with pytest.raises(ValueError) as err:
    #         V3Local.from_paserk(paserk, unsealing_key=token_bytes(32))
    #         pytest.fail("Key.from_paserk should fail.")
    #     assert msg in str(err.value)

    # def test_v3_local_from_paserk_with_wrong_unsealing_key(self):

    #     k = Key.new(3, "local", token_bytes(32))
    #     with open(get_path("keys/public_key_x25519.pem")) as key_file:
    #         sealed_key = k.to_paserk(sealing_key=key_file.read())

    #     with open(get_path("keys/private_key_x25519_2.pem")) as key_file:
    #         unsealing_key = key_file.read()

    #     with pytest.raises(DecryptError) as err:
    #         Key.from_paserk(sealed_key, unsealing_key=unsealing_key)
    #         pytest.fail("Key.from_paserk should fail.")
    #     assert "Failed to unseal a key." in str(err.value)


class TestV3Public:
    """
    Tests for v3.public.
    """

    def test_v3_public_to_paserk_id(self):
        sk = Key.new(3, "public", load_key("keys/private_key_ecdsa_p384.pem"))
        pk = Key.new(3, "public", load_key("keys/public_key_ecdsa_p384.pem"))
        assert sk.to_peer_paserk_id() == ""
        assert pk.to_peer_paserk_id() == ""

    def test_v3_public_verify_via_encode_with_wrong_key(self):
        sk = Key.new(3, "public", load_key("keys/private_key_ecdsa_p384.pem"))
        pk = Key.new(3, "public", load_key("keys/public_key_ecdsa_p384_2.pem"))
        token = pyseto.encode(sk, b"Hello world!")
        with pytest.raises(VerifyError) as err:
            pyseto.decode(pk, token)
            pytest.fail("pyseto.decode() should fail.")
        assert "Failed to verify." in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.public.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k4.public.AAAAAAAAAAAAAAAA", "Invalid PASERK version: k4."),
            ("k3.public.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k3.local.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k3.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
            ("k3.local.AAAAAAAAAAAAAAAA", "Invalid PASERK type: local."),
            (
                "k3.local-wrap.AAAAAAAAAAAAAAAA",
                "Invalid PASERK type: local-wrap.",
            ),
            (
                "k3.secret-wrap.AAAAAAAAAAAAAAAA",
                "secret-wrap needs wrapping_key.",
            ),
        ],
    )
    def test_v3_public_from_paserk_with_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V3Public.from_paserk(paserk)
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.secret-wrap.pie.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k3.secret-wrap.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k3.secret-wrap.xxx.AAAAAAAAAAAAAAAA", "Unknown wrapping algorithm: xxx."),
            ("k3.xxx.pie.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
        ],
    )
    def test_v3_public_from_paserk_with_wrapping_key_and_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V3Public.from_paserk(paserk, wrapping_key=token_bytes(32))
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    def test_v3_public_from_public_bytes_with_invalid_args(self):

        with pytest.raises(ValueError) as err:
            V3Public.from_public_bytes(b"xxx")
            pytest.fail("Key.from_paserk should fail.")
        assert "Invalid bytes for the key." in str(err.value)

    def test_v3_public_sign_via_encode_with_invalid_key(self):

        k = Key.from_paserk(
            "k3.secret-pw.mXsR2qVqmcDxmSWeQCnCwNeIxe5RDQ3ehnQvdXFj-YgAAAPoFI8eRXCL8PFpVW_CWOvGHnvMPy0BkMlKF1AtmBYGKold9i-ALC2oflkemYdbncrHbiKGd8zfjTQu2tTo2ayOMHybk_-hhopwJ2IUallYfLfUzPuqvtOQfVxXLtUBPnmR75dhRiPDgzdIO1OMbqa3Z1LDevvzbrcPyhHqmJSZioeJ7j1Mu8DJOvrIK0pWHmjDq_eg4YFnaOgz7I3Tkxx89A",
            password="correct horse battery staple".encode("utf-8").hex(),
        )
        with pytest.raises(SignError) as err:
            pyseto.encode(k, b"Hello world!")
            pytest.fail("pyseto.sign() should fail.")
        assert "Failed to sign." in str(err.value)



import json

import pytest
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_pem_private_key,
)

import pyseto
from pyseto import Key
from pyseto.versions.v3 import V3Public

from .utils import get_path


def _load_tests(paths: list) -> list:

    tests: list = []
    for path in paths:
        with open(get_path(path)) as tv_file:
            tv = json.loads(tv_file.read())
        tests += tv["tests"]
    return tests


def _name_to_version(name: str) -> int:

    v = name.split(".")[0]
    if len(v) != 2:
        raise ValueError("Invalid PASERK test name.")
    return int(v[1:])


def _public_key_compress(x: int, y: int) -> bytes:

    bx = x.to_bytes(48, byteorder="big")
    by = y.to_bytes((y.bit_length() + 7) // 8, byteorder="big")
    s = bytearray(1)
    s[0] = 0x02 + (by[len(by) - 1] & 1)
    return bytes(s) + bx


def _from_Ed25519PrivateKey_to_X25519PrivateKey(data: bytes) -> X25519PrivateKey:

    hasher = hashes.Hash(hashes.SHA512())
    hasher.update(data)
    h = bytearray(hasher.finalize())

    # curve25519 clamping
    h[0] &= 248
    h[31] &= 127
    h[31] |= 64
    return bytes(h[0:32])


class TestWithTestVectors:
    """
    Tests with test vectors defined in https://github.com/paseto-standard/test-vectors.
    """

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/v1.json",
                "vectors/v2.json",
                "vectors/v3.json",
                "vectors/v4.json",
            ]
        ),
    )
    def test_with_test_vectors(self, v):

        token = v["token"].encode("utf-8")
        payload = v["payload"]
        footer = v["footer"].encode("utf-8")
        implicit_assertion = v["implicit-assertion"].encode("utf-8")

        version = int(v["name"].split("-")[0])
        purpose = v["name"].split("-")[1]

        if v["expect-fail"]:
            if "public-key" not in v:
                nonce = bytes.fromhex(v["nonce"])
                key = bytes.fromhex(v["key"])

                k = Key.new(version, "local", key=key)
                with pytest.raises(ValueError) as err:
                    pyseto.encode(k, payload, footer, implicit_assertion, nonce=nonce)
                    pytest.fail("encode should fail.")
                assert "payload should be bytes, str or dict." in str(err.value)
                return

            secret_key_pem = v["secret-key"] if version == 1 else v["secret-key-pem"]
            public_key_pem = v["public-key"] if version == 1 else v["public-key-pem"]

            sk = Key.new(version, "public", secret_key_pem)
            with pytest.raises(ValueError) as err:
                pyseto.encode(sk, payload, footer, implicit_assertion)
                pytest.fail("encode should fail.")
            assert "payload should be bytes, str or dict." in str(err.value)
            return

        payload = payload.encode("utf-8")
        if purpose == "E":
            nonce = bytes.fromhex(v["nonce"])
            key = bytes.fromhex(v["key"])

            k = Key.new(version, "local", key=key)
            encoded = pyseto.encode(k, payload, footer, implicit_assertion, nonce=nonce)
            decoded_token = pyseto.decode(k, token, implicit_assertion)
            decoded = pyseto.decode(k, encoded, implicit_assertion)
            assert payload == decoded_token.payload == decoded.payload
            return

        if purpose == "S":
            secret_key_pem = v["secret-key"] if version == 1 else v["secret-key-pem"]
            public_key_pem = v["public-key"] if version == 1 else v["public-key-pem"]

            sk = Key.new(version, "public", secret_key_pem)
            encoded = pyseto.encode(sk, payload, footer, implicit_assertion)
            pk = Key.new(version, "public", public_key_pem)
            decoded_token = pyseto.decode(pk, token, implicit_assertion)
            decoded = pyseto.decode(pk, encoded, implicit_assertion)
            assert payload == decoded_token.payload == decoded.payload

            if version == 1:
                return

            secret_key = bytes.fromhex(v["secret-key"])
            public_key = bytes.fromhex(v["public-key"])

            if version == 3:
                # TODO add support for secret-key/public-key on v3.public test vectors.
                return

            sk = Key.from_asymmetric_key_params(version, d=secret_key[0:32])
            encoded = pyseto.encode(sk, payload, footer, implicit_assertion)
            pk = Key.from_asymmetric_key_params(version, x=public_key)
            decoded_token = pyseto.decode(pk, token, implicit_assertion)
            decoded = pyseto.decode(pk, encoded, implicit_assertion)
            assert payload == decoded_token.payload == decoded.payload
            return
        pytest.fail(f"Invalid test name: {v['name']}")

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.public.json",
                "vectors/PASERK/k2.public.json",
                "vectors/PASERK/k3.public.json",
                "vectors/PASERK/k4.public.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_public(self, v):

        version = _name_to_version(v["name"])
        if version == 1:
            k = Key.new(version, "public", v["key"])
        elif version == 2 or version == 4:
            k = Key.from_asymmetric_key_params(version, x=bytes.fromhex(v["key"]))
        elif version == 3:
            k = V3Public.from_public_bytes(bytes.fromhex(v["key"]))
        else:
            pytest.fail("Unsupported version.")
        assert k.to_paserk() == v["paserk"]
        k2 = Key.from_paserk(v["paserk"])
        assert k2.to_paserk() == v["paserk"]

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.secret.json",
                "vectors/PASERK/k2.secret.json",
                "vectors/PASERK/k3.secret.json",
                "vectors/PASERK/k4.secret.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_secret(self, v):

        version = _name_to_version(v["name"])
        if version == 1:
            k = Key.new(version, "public", v["key"])
        elif version == 2 or version == 4:
            k = Key.from_asymmetric_key_params(version, d=bytes.fromhex(v["secret-key-seed"]))
        elif version == 3:
            pub_k = Key.new(version, "public", bytes.fromhex(v["public-key"]))
            bx = pub_k._key.public_numbers().x.to_bytes(48, byteorder="big")
            by = pub_k._key.public_numbers().y.to_bytes(48, byteorder="big")
            k = Key.from_asymmetric_key_params(version, x=bx, y=by, d=bytes.fromhex(v["key"]))
        else:
            pytest.fail("Unsupported version.")
        assert k.to_paserk() == v["paserk"]
        k2 = Key.from_paserk(v["paserk"])
        assert k2.to_paserk() == v["paserk"]

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.local.json",
                "vectors/PASERK/k2.local.json",
                "vectors/PASERK/k3.local.json",
                "vectors/PASERK/k4.local.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_local(self, v):

        version = _name_to_version(v["name"])
        k = Key.new(version, "local", bytes.fromhex(v["key"]))
        k2 = Key.from_paserk(v["paserk"])
        assert k.to_paserk() == v["paserk"]
        assert k2.to_paserk() == v["paserk"]

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.pid.json",
                "vectors/PASERK/k2.pid.json",
                "vectors/PASERK/k3.pid.json",
                "vectors/PASERK/k4.pid.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_pid(self, v):

        version = _name_to_version(v["name"])
        if version == 1:
            k = Key.new(version, "public", v["key"])
        elif version == 2 or version == 4:
            k = Key.from_asymmetric_key_params(version, x=bytes.fromhex(v["key"]))
        elif version == 3:
            k = V3Public.from_public_bytes(bytes.fromhex(v["key"]))
        else:
            pytest.fail("Unsupported version.")
        assert k.to_paserk_id() == v["paserk"]

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.sid.json",
                "vectors/PASERK/k2.sid.json",
                "vectors/PASERK/k3.sid.json",
                "vectors/PASERK/k4.sid.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_sid(self, v):

        version = _name_to_version(v["name"])
        if version == 1:
            k = Key.new(version, "public", v["key"])
        elif version == 2 or version == 4:
            k = Key.from_asymmetric_key_params(version, d=bytes.fromhex(v["seed"]))
        elif version == 3:
            pub_k = Key.new(version, "public", bytes.fromhex(v["public-key"]))
            bx = pub_k._key.public_numbers().x.to_bytes(48, byteorder="big")
            by = pub_k._key.public_numbers().y.to_bytes(48, byteorder="big")
            k = Key.from_asymmetric_key_params(version, x=bx, y=by, d=bytes.fromhex(v["key"]))
        else:
            pytest.fail("Unsupported version.")
        assert k.to_paserk_id() == v["paserk"]

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.lid.json",
                "vectors/PASERK/k2.lid.json",
                "vectors/PASERK/k3.lid.json",
                "vectors/PASERK/k4.lid.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_lid(self, v):

        version = _name_to_version(v["name"])
        k = Key.new(version, "local", bytes.fromhex(v["key"]))
        assert k.to_paserk_id() == v["paserk"]

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.local-wrap.pie.json",
                "vectors/PASERK/k2.local-wrap.pie.json",
                "vectors/PASERK/k3.local-wrap.pie.json",
                "vectors/PASERK/k4.local-wrap.pie.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_local_wrap_pie(self, v):

        version = _name_to_version(v["name"])
        k = Key.from_paserk(v["paserk"], wrapping_key=bytes.fromhex(v["wrapping-key"]))

        k1 = Key.new(version, "local", bytes.fromhex(v["unwrapped"]))
        wpk = k1.to_paserk(wrapping_key=bytes.fromhex(v["wrapping-key"]))
        k2 = Key.from_paserk(wpk, wrapping_key=bytes.fromhex(v["wrapping-key"]))

        t = pyseto.encode(k, b"Hello world!")
        d = pyseto.decode(k, t)
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d.payload == d1.payload == d2.payload == b"Hello world!"

        t = pyseto.encode(k1, b"Hello world!")
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d1.payload == d2.payload == b"Hello world!"

        d = pyseto.decode(k, t)
        assert d.payload == b"Hello world!"

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.secret-wrap.pie.json",
                "vectors/PASERK/k2.secret-wrap.pie.json",
                "vectors/PASERK/k3.secret-wrap.pie.json",
                "vectors/PASERK/k4.secret-wrap.pie.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_secret_wrap_pie(self, v):

        version = _name_to_version(v["name"])

        k = Key.from_paserk(v["paserk"], wrapping_key=bytes.fromhex(v["wrapping-key"]))

        if version == 1:
            k1 = Key.new(version, "public", v["unwrapped"])
        elif version == 2 or version == 4:
            k1 = Key.from_asymmetric_key_params(version, d=bytes.fromhex(v["unwrapped"])[0:32])
        elif version == 3:
            pub_k = Key.new(version, "public", bytes.fromhex(v["public-key"]))
            bx = pub_k._key.public_numbers().x.to_bytes(48, byteorder="big")
            by = pub_k._key.public_numbers().y.to_bytes(48, byteorder="big")
            k1 = Key.from_asymmetric_key_params(version, x=bx, y=by, d=bytes.fromhex(v["unwrapped"]))
        else:
            pytest.fail("Unsupported version.")

        wpk = k1.to_paserk(wrapping_key=bytes.fromhex(v["wrapping-key"]))
        k2 = Key.from_paserk(wpk, wrapping_key=bytes.fromhex(v["wrapping-key"]))

        t = pyseto.encode(k, b"Hello world!")
        d = pyseto.decode(k, t)
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d.payload == d1.payload == d2.payload == b"Hello world!"

        t = pyseto.encode(k1, b"Hello world!")
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d1.payload == d2.payload == b"Hello world!"

        d = pyseto.decode(k, t)
        assert d.payload == b"Hello world!"

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.local-pw.json",
                "vectors/PASERK/k2.local-pw.json",
                "vectors/PASERK/k3.local-pw.json",
                "vectors/PASERK/k4.local-pw.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_local_pw(self, v):

        version = _name_to_version(v["name"])

        k = Key.from_paserk(v["paserk"], password=v["password"])

        k1 = Key.new(version, "local", bytes.fromhex(v["unwrapped"]))
        if version in [1, 3]:
            wpk = k1.to_paserk(password=v["password"], iteration=v["options"]["iterations"])
        elif version in [2, 4]:
            wpk = k1.to_paserk(
                password=v["password"],
                memory_cost=int(v["options"]["memlimit"] / 1024),
                time_cost=v["options"]["opslimit"],
            )
        else:
            pytest.fail("Unsupported version.")

        k2 = Key.from_paserk(wpk, password=v["password"])
        assert k1._key == k2._key

        t = pyseto.encode(k, b"Hello world!")
        d = pyseto.decode(k, t)
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d.payload == d1.payload == d2.payload == b"Hello world!"

        t = pyseto.encode(k1, b"Hello world!")
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d1.payload == d2.payload == b"Hello world!"

        d = pyseto.decode(k, t)
        assert d.payload == b"Hello world!"
        version = _name_to_version(v["name"])

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k1.secret-pw.json",
                "vectors/PASERK/k2.secret-pw.json",
                "vectors/PASERK/k3.secret-pw.json",
                "vectors/PASERK/k4.secret-pw.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_secret_pw(self, v):

        version = _name_to_version(v["name"])

        k = Key.from_paserk(v["paserk"], password=v["password"])

        if version == 1:
            k1 = Key.new(version, "public", v["unwrapped"])
        elif version == 2 or version == 4:
            k1 = Key.from_asymmetric_key_params(version, d=bytes.fromhex(v["unwrapped"])[0:32])
        elif version == 3:
            pub_k = Key.new(version, "public", bytes.fromhex(v["public-key"]))
            bx = pub_k._key.public_numbers().x.to_bytes(48, byteorder="big")
            by = pub_k._key.public_numbers().y.to_bytes(48, byteorder="big")
            k1 = Key.from_asymmetric_key_params(version, x=bx, y=by, d=bytes.fromhex(v["unwrapped"]))
        else:
            pytest.fail("Unsupported version.")

        if version in [1, 3]:
            wpk = k1.to_paserk(password=v["password"], iteration=v["options"]["iterations"])
        elif version in [2, 4]:
            wpk = k1.to_paserk(
                password=v["password"],
                memory_cost=int(v["options"]["memlimit"] / 1024),
                time_cost=v["options"]["opslimit"],
            )
        else:
            pytest.fail("Unsupported version.")

        k2 = Key.from_paserk(wpk, password=v["password"])

        t = pyseto.encode(k, b"Hello world!")
        d = pyseto.decode(k, t)
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d.payload == d1.payload == d2.payload == b"Hello world!"

        t = pyseto.encode(k1, b"Hello world!")
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d1.payload == d2.payload == b"Hello world!"

        d = pyseto.decode(k, t)
        assert d.payload == b"Hello world!"

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/PASERK/k2.seal.json",
                "vectors/PASERK/k4.seal.json",
            ]
        ),
    )
    def test_with_test_vectors_paserk_seal_v2_v4(self, v):

        version = _name_to_version(v["name"])

        if version == 1:
            raise NotImplementedError("Not implemented.")
        elif version in [2, 4]:
            sk_ed25519 = bytes.fromhex(v["sealing-secret-key"])[0:32]
            tmp_key = _from_Ed25519PrivateKey_to_X25519PrivateKey(sk_ed25519)
            sk_x25519 = X25519PrivateKey.from_private_bytes(tmp_key)
            unsealing_key = sk_x25519.private_bytes(
                Encoding.PEM,
                PrivateFormat.PKCS8,
                NoEncryption(),
            )
            sealing_key = sk_x25519.public_key().public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)
        elif version == 3:
            sk = load_pem_private_key(v["sealing-secret-key"].encode("utf-8"), password=None)
            unsealing_key = sk.private_numbers().private_value.to_bytes(48, byteorder="big")
            sealing_key = _public_key_compress(
                sk.private_numbers().public_numbers.x,
                sk.private_numbers().public_numbers.y,
            )
        else:
            pytest.fail("Unsupported version.")

        k = Key.from_paserk(v["paserk"], unsealing_key=unsealing_key)
        k1 = Key.new(version, "local", bytes.fromhex(v["unsealed"]))
        wpk = k1.to_paserk(sealing_key=sealing_key)

        k2 = Key.from_paserk(wpk, unsealing_key=unsealing_key)
        assert k1._key == k2._key

        t = pyseto.encode(k, b"Hello world!")
        d = pyseto.decode(k, t)
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d.payload == d1.payload == d2.payload == b"Hello world!"

        t = pyseto.encode(k1, b"Hello world!")
        d1 = pyseto.decode(k1, t)
        d2 = pyseto.decode(k2, t)
        assert d1.payload == d2.payload == b"Hello world!"

        d = pyseto.decode(k, t)
        assert d.payload == b"Hello world!"

    # @pytest.mark.parametrize(
    #     "v",
    #     _load_tests(
    #         [
    #             # "vectors/PASERK/k1.seal.json",
    #             "vectors/PASERK/k3.seal.json",
    #         ]
    #     ),
    # )
    # def test_with_test_vectors_paserk_seal_v1_v3(self, v):

    #     version = _name_to_version(v["name"])

    #     if version == 1:
    #         raise NotImplementedError("Not implemented.")
    #     elif version == 3:
    #         sk = load_pem_private_key(
    #             v["sealing-secret-key"].encode("utf-8"), password=None
    #         )
    #         unsealing_key = sk.private_numbers().private_value.to_bytes(
    #             48, byteorder="big"
    #         )
    #         sealing_key = _public_key_compress(
    #             sk.private_numbers().public_numbers.x,
    #             sk.private_numbers().public_numbers.y,
    #         )
    #     else:
    #         pytest.fail("Unsupported version.")

    #     # k = Key.from_paserk(v["paserk"], unsealing_key=unsealing_key)
    #     k1 = Key.new(version, "local", bytes.fromhex(v["unsealed"]))
    #     wpk = k1.to_paserk(sealing_key=sealing_key)

    #     k2 = Key.from_paserk(wpk, unsealing_key=unsealing_key)
    #     assert k1._key == k2._key

    #     # t = pyseto.encode(k, b"Hello world!")
    #     # d = pyseto.decode(k, t)
    #     # d1 = pyseto.decode(k1, t)
    #     # d2 = pyseto.decode(k2, t)
    #     # assert d.payload == d1.payload == d2.payload == b"Hello world!"

    #     t = pyseto.encode(k1, b"Hello world!")
    #     d1 = pyseto.decode(k1, t)
    #     d2 = pyseto.decode(k2, t)
    #     assert d1.payload == d2.payload == b"Hello world!"

    #     # d = pyseto.decode(k, t)
    #     # assert d.payload == b"Hello world!"



import json
import os

from pyseto.utils import base64url_decode


def get_path(name: str) -> str:
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), name)


def load_key(name: str) -> str:
    with open(get_path(name)) as key_file:
        k = key_file.read()
    return k


def load_jwk(name: str) -> str:
    with open(get_path(name)) as key_file:
        jwk = json.loads(key_file.read())
    res = {}
    res["d"] = base64url_decode(jwk["d"]) if "d" in jwk else b""
    res["x"] = base64url_decode(jwk["x"]) if "x" in jwk else b""
    res["y"] = base64url_decode(jwk["y"]) if "y" in jwk else b""

    if "d" in jwk and "x" in jwk and "y" not in jwk:
        res["x"] = b""
    return res



import json
from datetime import datetime, timedelta, timezone
from typing import Any, List, Optional, Union

import iso8601

from .exceptions import VerifyError
from .key_interface import KeyInterface
from .token import Token
from .utils import base64url_encode


class Paseto(object):
    """
    A PASETO processor which can be used as a PASETO encoder/decoder.
    """

    def __init__(
        self,
        exp: int = 0,
        include_iat: bool = False,
        leeway: int = 0,
    ):
        self._exp = exp
        self._include_iat = include_iat
        self._leeway = leeway
        return

    @classmethod
    def new(
        cls,
        exp: int = 0,
        include_iat: bool = False,
        leeway: int = 0,
    ):
        """
        Constructor of PASETO processor.

        Args:
            exp (int): A default expiration time (seconds) of PASETO tokens.
                It will be set in the payload as the registered ``exp`` claim
                when calling ``encode()`` with serializer=``json`` and this
                value > ``0``. If the value <= ``0``, the ``exp`` claim will
                not be set. In addition, this value can be overwritten by the
                ``exp`` parameter of ``encode()``. The default value is ``0``.
            include_iat (bool): If this value is ``True``, PASETO tokens
                which are created through ``encode()`` include an ``iat`` claim
                when calling ``encode()`` with serializer=``json``. The default
                value is ``False``.
            leeway (int): The leeway in seconds for validating ``exp`` and
                ``nbf``. The default value is ``0``.
        Returns:
            bytes: A PASETO processor object.
        """
        return cls(exp, include_iat, leeway)

    def encode(
        self,
        key: KeyInterface,
        payload: Union[bytes, str, dict],
        footer: Union[bytes, str, dict] = b"",
        implicit_assertion: Union[bytes, str] = b"",
        nonce: bytes = b"",
        serializer: Any = json,
        exp: int = 0,
    ) -> bytes:

        """
        Encodes a message to a PASETO token with a key for encryption or signing.

        Args:
            key (KeyInterface): A key for encryption or signing.
            payload (Union[bytes, str, dict]): A message to be encrypted or signed.
            footer (Union[bytes, str, dict]): A footer.
            implicit_assertion (Union[bytes, str]): An implicit assertion. It is
                only used in ``v3`` or ``v4``.
            nonce (bytes): A nonce. If omitted(it's recommended), a nonce will be
                generated with ``secrets.token_bytes()`` internally. If you don't
                want ot use ``secrets.token_bytes()``, you can specify it via this
                parameter explicitly.
            serializer (Any): A serializer which is used when the type of
                ``payload`` and/or ``footer`` is ``dict``. It must have a
                ``dumps()`` function to serialize the payload. Typically, you
                can use ``json`` or ``cbor2``.
            exp (int): An expiration time (seconds) of the PASETO token. It will be
                set in the payload as the registered ``exp`` claim when serializer
                is ``json`` and this value > ``0``. If the value <= ``0``, the
                ``exp`` claim will not be set.
        Returns:
            bytes: A PASETO token.
        Raise:
            ValueError: Invalid arguments.
            EncryptError: Failed to encrypt the message.
            SignError: Failed to sign the message.
        """

        if not isinstance(payload, (bytes, str, dict)):
            raise ValueError("payload should be bytes, str or dict.")

        bp: bytes
        if isinstance(payload, dict):
            if not serializer:
                raise ValueError("serializer should be specified for the payload object.")
            try:
                if not callable(serializer.dumps):
                    raise ValueError("serializer should have dumps().")
            except AttributeError:
                raise ValueError("serializer should have dumps().")
            except Exception:
                raise
            try:
                payload = self._set_registered_claims(payload, exp)
                bp = serializer.dumps(payload).encode("utf-8")
            except Exception as err:
                raise ValueError("Failed to serialize the payload.") from err
        else:
            bp = payload if isinstance(payload, bytes) else payload.encode("utf-8")

        bf: bytes
        if isinstance(footer, dict):
            if not serializer:
                raise ValueError("serializer should be specified for the footer object.")
            try:
                if not callable(serializer.dumps):
                    raise ValueError("serializer should have dumps().")
            except AttributeError:
                raise ValueError("serializer should have dumps().")
            except Exception:
                raise
            try:
                bf = serializer.dumps(footer).encode("utf-8")
            except Exception as err:
                raise ValueError("Failed to serialize the footer.") from err
        else:
            bf = footer if isinstance(footer, bytes) else footer.encode("utf-8")

        bi = implicit_assertion if isinstance(implicit_assertion, bytes) else implicit_assertion.encode("utf-8")

        if key.purpose == "local":
            return key.encrypt(bp, bf, bi, nonce)

        sig = key.sign(bp, bf, bi)
        token = key.header + base64url_encode(bp + sig)
        if bf:
            token += b"." + base64url_encode(bf)
        return token

    def decode(
        self,
        keys: Union[KeyInterface, List[KeyInterface]],
        token: Union[bytes, str],
        implicit_assertion: Union[bytes, str] = b"",
        deserializer: Optional[Any] = None,
        aud: str = "",
    ) -> Token:

        """
        Decodes a PASETO token with a key for decryption and/or verifying.

        Args:
            keys (KeyInterface): A key for decryption or verifying the signature in the token.
            token (Union[bytes, str]): A PASETO token to be decrypted or verified.
            implicit_assertion (Union[bytes, str]): An implicit assertion. It is
                only used in ``v3`` or ``v4``.
            deserializer (Optional[Any]): A deserializer which is used when you want to
                deserialize a ``payload`` attribute in the response object. It must have a
                ``loads()`` function to deserialize the payload. Typically, you can use
                ``json`` or ``cbor2``.
            aud (str): An audience claim value for the token verification.
                If ``deserializer=json`` and the payload of the token does not
                include an ``aud`` value that matches this value, the
                verification will fail.
        Returns:
            Token: A parsed PASETO token object.
        Raise:
            ValueError: Invalid arguments.
            DecryptError: Failed to decrypt the message.
            VerifyError: Failed to verify the message.
        """

        if deserializer:
            try:
                if not callable(deserializer.loads):
                    raise ValueError("deserializer should have loads().")
            except AttributeError:
                raise ValueError("deserializer should have loads().")
            except Exception:
                raise

        keys = keys if isinstance(keys, list) else [keys]
        bi = implicit_assertion if isinstance(implicit_assertion, bytes) else implicit_assertion.encode("utf-8")

        failed = None
        t = Token.new(token)
        for k in keys:
            if k.header != t.header:
                continue
            try:
                if k.purpose == "local":
                    t.payload = k.decrypt(t.payload, t.footer, bi)
                    return t
                t.payload = k.verify(t.payload, t.footer, bi)
                try:
                    if deserializer:
                        t.payload = deserializer.loads(t.payload)
                except Exception as err:
                    raise ValueError("Failed to deserialize the payload.") from err
                if deserializer:
                    try:
                        if t.footer:
                            t.footer = deserializer.loads(t.footer)
                    except Exception:
                        pass
                    self._verify_registered_claims(t.payload, aud)
                return t
            except Exception as err:
                failed = err
        if failed:
            raise failed
        raise ValueError("key is not found for verifying the token.")

    def _set_registered_claims(self, claims: dict, exp: int) -> dict:

        now = datetime.now(tz=timezone.utc)
        # exp
        if exp > 0:
            claims["exp"] = (now + timedelta(seconds=exp)).isoformat(timespec="seconds")
        elif self._exp > 0:
            claims["exp"] = (now + timedelta(seconds=self._exp)).isoformat(timespec="seconds")
        # iat
        if self._include_iat:
            claims["iat"] = now.isoformat(timespec="seconds")
        return claims

    def _verify_registered_claims(self, claims: dict, aud: str):

        now = iso8601.parse_date(datetime.now(tz=timezone.utc).isoformat(timespec="seconds"))
        # In Python 3.7 or later, the following code can be used:
        # now = datetime.fromisoformat(
        #     datetime.now(tz=timezone.utc).isoformat(timespec="seconds")
        # )
        if "exp" in claims:
            try:
                exp = iso8601.parse_date(claims["exp"])
            except Exception as err:
                raise VerifyError("Invalid exp.") from err
            if now > exp + timedelta(seconds=self._leeway):
                raise VerifyError("Token expired.")
        if "nbf" in claims:
            try:
                nbf = iso8601.parse_date(claims["nbf"])
            except Exception as err:
                raise VerifyError("Invalid nbf.") from err
            if now < nbf - timedelta(seconds=self._leeway):
                raise VerifyError("Token has not been activated yet.")
        if aud:
            if "aud" not in claims or aud != claims["aud"]:
                raise VerifyError("aud verification failed.")
        return



import hashlib
import hmac
from secrets import token_bytes
from typing import Any, Union

from cryptography.hazmat.primitives import hashes

# from cryptography.hazmat.primitives.asymmetric import ec
# from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

# from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from .exceptions import DecryptError, EncryptError
from .key_interface import KeyInterface
from .utils import base64url_decode, base64url_encode

# from .utils import base64url_decode, base64url_encode, ec_public_key_compress


class NISTKey(KeyInterface):
    """
    NIST (v1, v3) PASETO key.
    """

    # _VERSION = 1 or 3
    # _TYPE = "local", "public" or "secret"

    def __init__(self, key: Any):

        super().__init__(self._VERSION, self._TYPE, key)
        return

    @classmethod
    def from_paserk(
        cls,
        paserk: str,
        wrapping_key: bytes = b"",
        password: bytes = b"",
        unsealing_key: bytes = b"",
    ) -> KeyInterface:

        if wrapping_key and password:
            raise ValueError("Only one of wrapping_key or password should be specified.")

        frags = paserk.split(".")
        if frags[0] != f"k{cls._VERSION}":
            raise ValueError(f"Invalid PASERK version: {frags[0]}.")

        if wrapping_key:
            # local-wrap
            if len(frags) != 4:
                raise ValueError("Invalid PASERK format.")
            if frags[2] != "pie":
                raise ValueError(f"Unknown wrapping algorithm: {frags[2]}.")

            h = frags[0] + "." + frags[1] + ".pie."
            if frags[1] == "local-wrap":
                return cls(cls._decode_pie(h, wrapping_key, frags[3]))
            raise ValueError(f"Invalid PASERK type: {frags[1]}.")

        if len(frags) != 3:
            raise ValueError("Invalid PASERK format.")

        if password:
            # local-pw
            h = frags[0] + "." + frags[1] + "."
            if frags[1] == "local-pw":
                return cls(cls._decode_pbkw(h, password, frags[2]))
            raise ValueError(f"Invalid PASERK type: {frags[1]}.")

        # if unsealing_key:
        #     # seal
        #     h = frags[0] + "." + frags[1] + "."
        #     if frags[1] == "seal":
        #         return cls(cls._decode_pke(h, unsealing_key, frags[2]))
        #     raise ValueError(f"Invalid PASERK type: {frags[1]}.")

        # local
        k = base64url_decode(frags[2])
        if frags[1] == "local":
            return cls(k)
        if frags[1] == "local-wrap":
            raise ValueError(f"{frags[1]} needs wrapping_key.")
        if frags[1] == "local-pw":
            raise ValueError(f"{frags[1]} needs password.")
        raise ValueError(f"Invalid PASERK type: {frags[1]}.")

    def to_paserk(
        self,
        wrapping_key: Union[bytes, str] = b"",
        password: Union[bytes, str] = b"",
        sealing_key: Union[bytes, str] = b"",
        iteration: int = 100000,
        memory_cost: int = 15 * 1024,
        time_cost: int = 2,
        parallelism: int = 1,
    ) -> str:

        if wrapping_key and password:
            raise ValueError("Only one of wrapping_key or password should be specified.")

        # local-wrap
        if wrapping_key:
            bkey = wrapping_key if isinstance(wrapping_key, bytes) else wrapping_key.encode("utf-8")
            h = f"k{self.version}.local-wrap.pie."
            return h + self._encode_pie(h, bkey, self._key)

        # seal
        # if sealing_key:
        #     bkey = (
        #         sealing_key
        #         if isinstance(sealing_key, bytes)
        #         else sealing_key.encode("utf-8")
        #     )
        #     h = f"k{self.version}.seal."
        #     return h + self._encode_pke(h, bkey, self._key)

        # local-pw
        if password:
            bpw = password if isinstance(password, bytes) else password.encode("utf-8")
            h = f"k{self.version}.local-pw."
            return h + self._encode_pbkw(h, bpw, self._key, iteration)

        # local
        h = f"k{self.version}.local."
        return h + base64url_encode(self._key).decode("utf-8")

    @classmethod
    def _encode_pie(cls, header: str, wrapping_key: bytes, ptk: bytes) -> str:

        h = header.encode("utf-8")
        n = token_bytes(32)
        x = cls._generate_hash(wrapping_key, b"\x80" + n)
        ek = x[0:32]
        n2 = x[32:]
        ak = cls._generate_hash(wrapping_key, b"\x81" + n, 32)
        c = cls._encrypt(ek, n2, ptk)
        t = cls._generate_hash(ak, h + n + c, 48)
        return base64url_encode(t + n + c).decode("utf-8")

    @classmethod
    def _decode_pie(cls, header: str, wrapping_key: bytes, data: str) -> bytes:

        h = header.encode("utf-8")
        d = base64url_decode(data)
        t = d[0:48]
        n = d[48:80]
        c = d[80:]
        ak = cls._generate_hash(wrapping_key, b"\x81" + n, 32)
        t2 = cls._generate_hash(ak, h + n + c, 48)
        if t != t2:
            raise DecryptError("Failed to unwrap a key.")
        x = cls._generate_hash(wrapping_key, b"\x80" + n)
        ek = x[0:32]
        n2 = x[32:]
        return cls._decrypt(ek, n2, c)

    @classmethod
    def _encode_pbkw(cls, header: str, password: bytes, ptk: bytes, iteration: int = 100000) -> str:

        h = header.encode("utf-8")

        s = token_bytes(32)
        k = PBKDF2HMAC(
            algorithm=hashes.SHA384(),
            length=32,
            salt=s,
            iterations=iteration,
        ).derive(password)
        ek = cls._digest(b"\xff" + k)[0:32]
        ak = cls._digest(b"\xfe" + k)
        n = token_bytes(16)
        edk = cls._encrypt(ek, n, ptk)
        bi = iteration.to_bytes(4, byteorder="big")
        t = cls._generate_hash(ak, h + s + bi + n + edk, 48)
        return base64url_encode(s + bi + n + edk + t).decode("utf-8")

    @classmethod
    def _decode_pbkw(cls, header: str, password: bytes, data: str) -> bytes:

        h = header.encode("utf-8")
        d = base64url_decode(data)

        s = d[0:32]
        bi = d[32:36]
        n = d[36:52]
        edk = d[52 : len(d) - 48]
        t = d[-48:]
        i = int.from_bytes(bi, byteorder="big")

        k = PBKDF2HMAC(
            algorithm=hashes.SHA384(),
            length=32,
            salt=s,
            iterations=i,
        ).derive(password)

        ak = cls._digest(b"\xfe" + k)
        t2 = cls._generate_hash(ak, h + s + bi + n + edk, 48)
        if t != t2:
            raise DecryptError("Failed to unwrap a key.")

        ek = cls._digest(b"\xff" + k)[0:32]
        return cls._decrypt(ek, n, edk)

    # @classmethod
    # def _encode_pke(cls, header: str, sealing_key: bytes, ptk: bytes) -> str:

    #     h = header.encode("utf-8")

    #     esk = ec.generate_private_key(ec.SECP384R1())
    #     epk = ec_public_key_compress(
    #         esk.private_numbers().public_numbers.x,
    #         esk.private_numbers().public_numbers.y,
    #     )
    #     pub = EllipticCurvePublicKey.from_encoded_point(ec.SECP384R1(), sealing_key)
    #     pk = ec_public_key_compress(
    #         pub.public_numbers().x,
    #         pub.public_numbers().y,
    #     )
    #     xk = esk.exchange(ec.ECDH(), pub)
    #     tmp = cls._digest(b"\x01" + h + xk + epk + pk)
    #     ek = tmp[0:32]
    #     n = tmp[32:]
    #     ak = cls._digest(b"\x01" + h + xk + epk + pk)
    #     edk = cls._encrypt(ek, n, ptk)
    #     t = cls._generate_hash(ak, h + epk + edk, 48)
    #     return base64url_encode(t + epk + edk).decode("utf-8")

    # @classmethod
    # def _decode_pke(cls, header: str, unsealing_key: bytes, data: str) -> bytes:

    #     h = header.encode("utf-8")
    #     d = base64url_decode(data)

    #     if cls._VERSION == 1:
    #         raise NotImplementedError("Not implemented.")

    #     t = d[0:48]
    #     epk = d[48:97]
    #     edk = d[97:]
    #     sk = ec.derive_private_key(
    #         int.from_bytes(unsealing_key, byteorder="big"), ec.SECP384R1()
    #     )
    #     pub = EllipticCurvePublicKey.from_encoded_point(ec.SECP384R1(), epk)

    #     xk = sk.exchange(ec.ECDH(), pub)
    #     # shared = sk.exchange(ec.ECDH(), pub)
    #     # xk = HKDF(algorithm=hashes.SHA384(), length=48, salt=None, info=b"").derive(shared)
    #     pk = ec_public_key_compress(
    #         sk.private_numbers().public_numbers.x,
    #         sk.private_numbers().public_numbers.y,
    #     )
    #     ak = cls._digest(b"\x01" + h + xk + epk + pk)
    #     t2 = cls._generate_hash(ak, h + epk + edk, 48)
    #     if t != t2:
    #         raise DecryptError("Failed to unseal a key.")

    #     tmp = cls._digest(b"\x01" + h + xk + epk + pk)
    #     ek = tmp[0:32]
    #     n = tmp[32:]
    #     return cls._decrypt(ek, n, edk)

    @staticmethod
    def _generate_hash(key: bytes, msg: bytes, size: int = 0) -> bytes:

        try:
            d = hmac.new(key, msg, hashlib.sha384).digest()
            return d[0:size] if size > 0 else d
        except Exception as err:
            raise EncryptError("Failed to generate hash.") from err

    @staticmethod
    def _digest(msg: bytes) -> bytes:
        return hashlib.sha384(msg).digest()

    @staticmethod
    def _encrypt(key: bytes, nonce: bytes, msg: bytes) -> bytes:
        try:
            encryptor = Cipher(algorithms.AES(key), modes.CTR(nonce)).encryptor()
            return encryptor.update(msg)
        except Exception as err:
            raise EncryptError("Failed to encrypt.") from err

    @staticmethod
    def _decrypt(key: bytes, nonce: bytes, msg: bytes) -> bytes:
        try:
            decryptor = Cipher(algorithms.AES(key), modes.CTR(nonce)).decryptor()
            return decryptor.update(msg)
        except Exception as err:
            raise DecryptError("Failed to decrypt.") from err



from typing import Any, Union

from .exceptions import NotSupportedError


class KeyInterface:
    """
    The key interface class for PASETO.

    :func:`pyseto.Key.new <pyseto.Key.new>` returns an object which has this interface.
    """

    _VERSION = 0
    _TYPE = ""

    def __init__(self, version: int, purpose: str, key: Any):
        self._version = version
        self._purpose = purpose
        self._header = (f"v{self._version}." + self._purpose + ".").encode("utf-8")
        self._sig_size = 0
        self._key: Any = key
        if not self._key:
            raise ValueError("key must be specified.")
        self._is_secret = True
        return

    @property
    def version(self) -> int:
        """
        The version of the key. It will be ``1``, ``2``, ``3`` or ``4``.
        """
        return self._version

    @property
    def purpose(self) -> str:
        """
        The purpose of the key. It will be ``"local"`` or ``"public"``.
        """
        return self._purpose

    @property
    def header(self) -> bytes:
        """
        The header value for a PASETO token. It will be ``"v<version>.<purpose>."``.
        For example, ``"v1.local."``.
        """
        return self._header

    @property
    def is_secret(self) -> bool:
        """
        If it is True, the key is a symmetric key or an asymmetric secret key.
        """
        return self._is_secret

    # @property
    # def key(self) -> Any:
    #     """
    #     Byte string or pyca/cryptography key object of the key.
    #     If the key is ``public``, the key object is one of the pyca/cryptography objects bellow:

    #     - ``RSAPrivateKey``
    #     - ``RSAPublicKey``
    #     - ``Ed25519PrivateKey``
    #     - ``Ed25519PublicKey``
    #     - ``EllipticCurvePrivateKey``
    #     - ``EllipticCurvePublicKey``
    #     """
    #     return self._key

    def to_paserk(
        self,
        wrapping_key: Union[bytes, str] = b"",
        password: Union[bytes, str] = b"",
        sealing_key: Union[bytes, str] = b"",
        iteration: int = 100000,
        memory_cost: int = 15 * 1024,
        time_cost: int = 2,
        parallelism: int = 1,
    ) -> str:
        """
        Returns the PASERK expression of the key.

        Args:
            wrapping_key (Union[bytes, str]): A wrapping key to wrap the key.
                If the `wrapping_key` is specified, `password` should not be
                specified.
            password (Union[bytes, str]): A password to wrap the key. If the
                `password` is specified, `wrapping_key` should not be specified.
            iteration (int): An iteration count used for password-based key
                wrapping. This argument will only be used when the `password` is
                specified.
            memory_cost (int): Amount of memory to use for password-based key
                wrapping using argon2. This argument will only be used when
                the `password` is specified for `v2/v4` key.
            time_cost (int):  Number of iterations to perform for password-based
                key wrapping using argon2. This argument will only be used when
                the `password` is specified for `v2/v4` key.
            parallelism (int): Degree of parallelism for password-based key
                wrapping using argon2. This argument will only be used when
                the `password` is specified for `v2/v4` key.
        Returns:
            str: A PASERK string.
        Raise:
            ValueError: Invalid arguments.
            EncryptError: Failed to wrap the key.
        """
        raise NotImplementedError("The PASERK expression for the key is not supported yet.")

    def to_paserk_id(self) -> str:
        """
        Returns the PASERK ID of the key.

        Returns:
            str: A PASERK ID string.
        """
        raise NotImplementedError("The PASERK ID for the key is not supported yet.")

    def to_peer_paserk_id(self) -> str:
        """
        Returns the peer(public) PASERK ID of the key.
        It can be used only in case that the key is `k2.secret` or `k4.secret`.

        Returns:
            str: A peer PASERK ID string. If the key is neither `k2.secret` nor
                `k4.secret`, an empty string will be returned.
        """
        return ""

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:
        """
        Encrypts a message to a PASETO token with the key.

        This function is calld in :func:`pyseto.encode <pyseto.encode>`  so you
        don't need to call it directly.

        Args:
            payload (bytes): A message to be encrypted which will be the payload
                part of the PASETO token.
            footer (bytes): A footer.
            implicit_assertion (Union[bytes, str]): An implicit assertion. It is
                only used in ``v3`` or ``v4``.
            nonce (bytes): A nonce.
        Returns:
            bytes: A PASETO token.
        Raise:
            ValueError: Invalid arguments.
            EncryptError: Failed to encrypt the message.
            NotSupportedError: The key does not support the operation.
        """
        raise NotSupportedError("A key for public does not have encrypt().")

    def decrypt(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:
        """
        Decrypts an encrypted PASETO token with the key.

        This function is calld in :func:`pyseto.decode <pyseto.decode>`  so you
        don't need to call it directly.

        Args:
            payload (bytes): A message to be decrypted which is the payload part
                of the PASETO token.
            footer (bytes): A footer.
            implicit_assertion (Union[bytes, str]): An implicit assertion. It is
                only used in ``v3`` or ``v4``.
        Returns:
            bytes: A dcrypted payload.
        Raise:
            ValueError: Invalid arguments.
            DecryptError: Failed to decrypt the message.
            NotSupportedError: The key does not support the operation.
        """
        raise NotSupportedError("A key for public does not have decrypt().")

    def sign(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:
        """
        Signs a message with the key and makes a PASETO token.

        This function is calld in :func:`pyseto.encode <pyseto.encode>`  so you
        don't need to call it directly.

        Args:
            payload (bytes): A message to be signed and encoded which will be
                the payload part of the PASETO token.
            footer (bytes): A footer.
            implicit_assertion (Union[bytes, str]): An implicit assertion. It is
                only used in ``v3`` or ``v4``.
            nonce (bytes): A nonce.
        Returns:
            bytes: A PASETO token.
        Raise:
            ValueError: Invalid arguments.
            EncryptError: Failed to sign the message.
            NotSupportedError: The key does not support the operation.
        """
        raise NotSupportedError("A key for local does not have sign().")

    def verify(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:
        """
        Verifies and decodes a signed PASETO token with the key.

        This function is calld in :func:`pyseto.decode <pyseto.decode>`  so you
        don't need to call it directly.

        Args:
            payload (bytes): A message to be verified and decoded which is the
                payload part of the PASETO token.
            footer (bytes): A footer.
            implicit_assertion (Union[bytes, str]): An implicit assertion. It is
                only used in ``v3`` or ``v4``.
        Returns:
            bytes: A verified and decoded payload.
        Raise:
            ValueError: Invalid arguments.
            DecryptError: Failed to verify the message.
            NotSupportedError: The key does not support the operation.
        """
        raise NotSupportedError("A key for local does not have verify().")



from secrets import token_bytes

import pytest

from pyseto import NotSupportedError
from pyseto.key_interface import KeyInterface


class TestKeyInterface:
    """
    Tests for KeyInterface.
    """

    @pytest.mark.parametrize(
        "version, purpose, key",
        [
            (1, "local", token_bytes(32)),
            (2, "local", token_bytes(32)),
            (3, "local", token_bytes(32)),
            (4, "local", token_bytes(32)),
        ],
    )
    def test_key_interface_constructor_local(self, version, purpose, key):
        k = KeyInterface(version, purpose, key)
        assert isinstance(k, KeyInterface)
        assert k.version == version
        assert k.purpose == purpose
        assert k.is_secret is True
        with pytest.raises(NotSupportedError) as err:
            k.encrypt(b"Hello world!")
            pytest.fail("KeyInterface.encrypt() should fail.")
        assert "A key for public does not have encrypt()." in str(err.value)
        with pytest.raises(NotSupportedError) as err:
            k.decrypt(b"xxxxxx")
            pytest.fail("KeyInterface.decrypt() should fail.")
        assert "A key for public does not have decrypt()." in str(err.value)
        with pytest.raises(NotSupportedError) as err:
            k.sign(b"Hello world!")
            pytest.fail("KeyInterface.sign() should fail.")
        assert "A key for local does not have sign()." in str(err.value)
        with pytest.raises(NotSupportedError) as err:
            k.verify(b"xxxxxx")
            pytest.fail("KeyInterface.verify() should fail.")
        assert "A key for local does not have verify()." in str(err.value)
        with pytest.raises(NotImplementedError) as err:
            k.to_paserk()
            pytest.fail("KeyInterface.to_paserk() should fail.")
        assert "The PASERK expression for the key is not supported yet." in str(err.value)
        with pytest.raises(NotImplementedError) as err:
            k.to_paserk_id()
            pytest.fail("KeyInterface.to_paserk_id() should fail.")
        assert "The PASERK ID for the key is not supported yet." in str(err.value)



import hashlib
from secrets import token_bytes
from typing import Any, Union

# from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305
from Cryptodome.Cipher import ChaCha20_Poly1305
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)

from ..exceptions import DecryptError, EncryptError, SignError, VerifyError
from ..key_sodium import SodiumKey
from ..utils import base64url_encode, pae


class V2Local(SodiumKey):
    """
    The key object for v2.local.
    """

    _VERSION = 2
    _TYPE = "local"

    def __init__(self, key: Union[str, bytes]):

        super().__init__(key)
        if len(self._key) != 32:
            raise ValueError("key must be 32 bytes long.")
        return

    def to_paserk_id(self) -> str:
        h = "k2.lid."
        p = self.to_paserk()
        b = hashlib.blake2b(digest_size=33)
        b.update((h + p).encode("utf-8"))
        d = b.digest()
        return h + base64url_encode(d).decode("utf-8")

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:

        n = self._generate_nonce(nonce, payload)
        pre_auth = pae([self.header, n, footer])

        try:
            cipher = ChaCha20_Poly1305.new(key=self._key, nonce=n)
            cipher.update(pre_auth)
            c, tag = cipher.encrypt_and_digest(payload)
            token = self._header + base64url_encode(n + c + tag)
            if footer:
                token += b"." + base64url_encode(footer)
            return token
        except Exception as err:
            raise EncryptError("Failed to encrypt.") from err

    def decrypt(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:

        n = payload[0:24]
        c = payload[24 : len(payload) - 16]
        tag = payload[-16:]
        pre_auth = pae([self.header, n, footer])

        try:
            cipher = ChaCha20_Poly1305.new(key=self._key, nonce=n)
            cipher.update(pre_auth)
            return cipher.decrypt_and_verify(c, tag)
        except Exception as err:
            raise DecryptError("Failed to decrypt.") from err

    @staticmethod
    def _generate_nonce(key: bytes, msg: bytes) -> bytes:

        if key:
            if len(key) != 24:
                raise ValueError("nonce must be 24 bytes long.")
        else:
            key = token_bytes(24)

        try:
            h = hashlib.blake2b(key=key, digest_size=24)
            h.update(msg)
            return h.digest()
        except Exception as err:
            raise EncryptError("Failed to generate internal nonce.") from err


class V2Public(SodiumKey):
    """
    The key object for v2.public.
    """

    _VERSION = 2
    _TYPE = "public"

    def __init__(self, key: Any):

        super().__init__(key)
        self._sig_size = 64

        if not isinstance(self._key, (Ed25519PublicKey, Ed25519PrivateKey)):
            raise ValueError("The key is not Ed25519 key.")

        if isinstance(self._key, Ed25519PublicKey):
            self._is_secret = False
        return

    def to_paserk_id(self) -> str:
        p = self.to_paserk()
        h = "k2.pid." if isinstance(self._key, Ed25519PublicKey) else "k2.sid."
        b = hashlib.blake2b(digest_size=33)
        b.update((h + p).encode("utf-8"))
        d = b.digest()
        return h + base64url_encode(d).decode("utf-8")

    # @classmethod
    # def from_public_bytes(cls, key: bytes):
    #     try:
    #         k = Ed25519PublicKey.from_public_bytes(key)
    #     except Exception as err:
    #         raise ValueError("Invalid bytes for the key.") from err
    #     return cls(k)

    def sign(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:

        if isinstance(self._key, Ed25519PublicKey):
            raise ValueError("A public key cannot be used for signing.")
        m2 = pae([self.header, payload, footer])
        try:
            return self._key.sign(m2)
        except Exception as err:
            raise SignError("Failed to sign.") from err

    def verify(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""):

        if len(payload) <= self._sig_size:
            raise ValueError("Invalid payload.")

        sig = payload[-self._sig_size :]
        m = payload[: len(payload) - self._sig_size]
        k = self._key if isinstance(self._key, Ed25519PublicKey) else self._key.public_key()
        m2 = pae([self.header, m, footer])
        try:
            k.verify(sig, m2)
        except Exception as err:
            raise VerifyError("Failed to verify.") from err
        return m



import json
import time
from datetime import datetime, timedelta, timezone
from secrets import token_bytes

import freezegun
import iso8601

import pyseto
from pyseto import Key, Paseto

from .utils import get_path


class TestSample:
    """
    Tests for sample code.
    """

    def test_sample_v4_local_old(self):

        key = Key.new(4, "local", b"our-secret")
        token = pyseto.encode(
            key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )

        decoded = pyseto.decode(key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_v4_public_old(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(4, "public", private_key_pem)
        token = pyseto.encode(
            private_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        public_key = Key.new(4, "public", public_key_pem)
        decoded = pyseto.decode(public_key, token)

        assert (
            token
            == b"v4.public.eyJkYXRhIjogInRoaXMgaXMgYSBzaWduZWQgbWVzc2FnZSIsICJleHAiOiAiMjAyMi0wMS0wMVQwMDowMDowMCswMDowMCJ9l1YiKei2FESvHBSGPkn70eFO1hv3tXH0jph1IfZyEfgm3t1DjkYqD5r4aHWZm1eZs_3_bZ9pBQlZGp0DPSdzDg"
        )
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_v4_local(self):

        key = Key.new(version=4, purpose="local", key=b"our-secret")
        token = pyseto.encode(
            key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )

        decoded = pyseto.decode(key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_v4_public(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token)

        assert (
            token
            == b"v4.public.eyJkYXRhIjogInRoaXMgaXMgYSBzaWduZWQgbWVzc2FnZSIsICJleHAiOiAiMjAyMi0wMS0wMVQwMDowMDowMCswMDowMCJ9l1YiKei2FESvHBSGPkn70eFO1hv3tXH0jph1IfZyEfgm3t1DjkYqD5r4aHWZm1eZs_3_bZ9pBQlZGp0DPSdzDg"
        )
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    @freezegun.freeze_time("2021-01-01")
    def test_sample_v4_public_with_serializer(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token, deserializer=json)

        assert (
            token
            == b"v4.public.eyJkYXRhIjogInRoaXMgaXMgYSBzaWduZWQgbWVzc2FnZSIsICJleHAiOiAiMjAyMi0wMS0wMVQwMDowMDowMCswMDowMCJ9l1YiKei2FESvHBSGPkn70eFO1hv3tXH0jph1IfZyEfgm3t1DjkYqD5r4aHWZm1eZs_3_bZ9pBQlZGp0DPSdzDg"
        )
        assert decoded.payload["data"] == "this is a signed message"
        assert decoded.payload["exp"] == "2022-01-01T00:00:00+00:00"

    def test_sample_v4_public_with_serializer_and_exp(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"
        now = datetime.now(tz=timezone.utc)

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message"},
            exp=3600,
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token, deserializer=json)

        assert decoded.payload["data"] == "this is a signed message"
        assert iso8601.parse_date(decoded.payload["exp"]) >= now + timedelta(seconds=3600 - 1)

    def test_sample_v4_public_with_paseto_class(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"
        now = datetime.now(tz=timezone.utc)

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        paseto = Paseto.new(exp=3600, include_iat=True)
        token = paseto.encode(
            private_key,
            {"data": "this is a signed message"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token, deserializer=json)

        assert decoded.payload["data"] == "this is a signed message"
        assert "iat" in decoded.payload
        assert "exp" in decoded.payload
        assert iso8601.parse_date(decoded.payload["exp"]) >= now + timedelta(seconds=3600 - 1)

    def test_sample_v4_public_with_paseto_class_and_leeway(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"
        now = datetime.now(tz=timezone.utc)

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        paseto = Paseto.new(exp=1, include_iat=True, leeway=1)
        token = paseto.encode(
            private_key,
            {"data": "this is a signed message"},
        )
        time.sleep(1)
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token, deserializer=json)

        assert decoded.payload["data"] == "this is a signed message"
        assert "iat" in decoded.payload
        assert "exp" in decoded.payload
        assert iso8601.parse_date(decoded.payload["exp"]) >= now

    def test_sample_v4_public_with_kid(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"
        now = datetime.now(tz=timezone.utc)

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        paseto = Paseto.new(exp=3600, include_iat=True)
        token = paseto.encode(
            private_key,
            {
                "data": "this is a signed message",
                "nbf": (now - timedelta(seconds=10)).isoformat(timespec="seconds"),
            },
            footer={"kid": public_key.to_paserk_id()},
        )
        decoded = pyseto.decode(public_key, token, deserializer=json)

        assert decoded.payload["data"] == "this is a signed message"
        assert "iat" in decoded.payload
        assert "exp" in decoded.payload
        assert "kid" in decoded.footer
        assert decoded.footer["kid"] == "k4.pid.yh4-bJYjOYAG6CWy0zsfPmpKylxS7uAWrxqVmBN2KAiJ"
        assert iso8601.parse_date(decoded.payload["exp"]) >= now

    def test_sample_paserk(self):

        symmetric_key = Key.new(version=4, purpose="local", key=b"our-secret")
        private_key = Key.from_paserk(
            "k4.secret.tMv7Q99M4hByfZU-SnEzB_oZu32fhQQUONnhG5QqN3Qeudu7vAR8A_1wYE4AcfCYfhayi3VyJcEfAEFdDiCxog"
        )
        public_key = Key.from_paserk("k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI")

        token = pyseto.encode(
            private_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        decoded = pyseto.decode(public_key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

        assert symmetric_key.to_paserk() == "k4.local.b3VyLXNlY3JldA"
        assert (
            private_key.to_paserk()
            == "k4.secret.tMv7Q99M4hByfZU-SnEzB_oZu32fhQQUONnhG5QqN3Qeudu7vAR8A_1wYE4AcfCYfhayi3VyJcEfAEFdDiCxog"
        )
        assert public_key.to_paserk() == "k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI"

    def test_sample_paserk_id(self):

        symmetric_key = Key.new(version=4, purpose="local", key=b"our-secret")
        private_key = Key.from_paserk(
            "k4.secret.tMv7Q99M4hByfZU-SnEzB_oZu32fhQQUONnhG5QqN3Qeudu7vAR8A_1wYE4AcfCYfhayi3VyJcEfAEFdDiCxog"
        )
        public_key = Key.from_paserk("k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI")

        assert symmetric_key.to_paserk_id() == "k4.lid._D6kgTzxgiPGk35gMj9bukgj4En2H94u22wVX9zaoh05"
        assert private_key.to_paserk_id() == "k4.sid.9gZFsAQuXhu9lif2pV3rCDjOewsMF4qb4RHGhc0zUklt"
        assert public_key.to_paserk_id() == "k4.pid.yh4-bJYjOYAG6CWy0zsfPmpKylxS7uAWrxqVmBN2KAiJ"

    def test_sample_paserk_key_wrapping_local(self):

        raw_key = Key.new(version=4, purpose="local", key=b"our-secret")
        wrapping_key = token_bytes(32)
        wpk = raw_key.to_paserk(wrapping_key=wrapping_key)

        # assert wpk == "k4.local-wrap.pie.TNKEwC4K1xBcgJ_GiwWAoRlQFE33HJO3oN9DHEZ05pieSCd-W7bgAL64VG9TZ_pBkuNBFHNrfOGHtnfnhYGdbz5-x3CxShhPJxg"

        unwrapped_key = Key.from_paserk(wpk, wrapping_key=wrapping_key)
        token = pyseto.encode(
            raw_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        decoded = pyseto.decode(unwrapped_key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_paserk_key_wrapping_public(self):

        raw_private_key = Key.from_paserk(
            "k4.secret.tMv7Q99M4hByfZU-SnEzB_oZu32fhQQUONnhG5QqN3Qeudu7vAR8A_1wYE4AcfCYfhayi3VyJcEfAEFdDiCxog"
        )
        public_key = Key.from_paserk("k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI")
        wrapping_key = token_bytes(32)
        wpk = raw_private_key.to_paserk(wrapping_key=wrapping_key)

        # assert wpk == "k4.secret-wrap.pie.excv7V4-NaECy5hpji-tkSkMvyjsAgNxA-mGALgdjyvGNyDlTb89bJ35R1e3tILgbMpEW5WXMXzySe2T-sBz-ZAcs1j7rbD3ZWvsBTM6K5N9wWfAxbR4ppCXH_H5__9yY-kBaF2NimyAJyduhOhSmqLm6TTSucpAOakEJOXePW8"

        unwrapped_private_key = Key.from_paserk(wpk, wrapping_key=wrapping_key)
        token = pyseto.encode(
            unwrapped_private_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        decoded = pyseto.decode(public_key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_paserk_password_local(self):

        raw_key = Key.new(version=4, purpose="local", key=b"our-secret")
        wpk = raw_key.to_paserk(password="our-secret")

        # assert wpk == "k4.local-pw.HrCs9Pu-2LB0l7jkHB-x2gAAAAAA8AAAAAAAAgAAAAGttW0IHZjQCHJdg-Vc3tqO_GSLR4vzLl-yrKk2I-l8YHj6jWpC0lQB2Z7uzTtVyV1rd_EZQPzHdw5VOtyucP0FkCU"

        unwrapped_key = Key.from_paserk(wpk, password="our-secret")
        token = pyseto.encode(
            raw_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        decoded = pyseto.decode(unwrapped_key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_paserk_password_public(self):

        raw_private_key = Key.from_paserk(
            "k4.secret.tMv7Q99M4hByfZU-SnEzB_oZu32fhQQUONnhG5QqN3Qeudu7vAR8A_1wYE4AcfCYfhayi3VyJcEfAEFdDiCxog"
        )
        public_key = Key.from_paserk("k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI")
        wpk = raw_private_key.to_paserk(password="our-secret")

        # assert wpk == "k4.secret-pw.MEMW4K1MaD5nWigCLyEyFAAAAAAA8AAAAAAAAgAAAAFU-tArtryNVjS2n2hCYiM11V6tOyuIog69Bjb0yNZanrLJ3afGclb3kPzQ6IhK8ob9E4QgRdEALGWCizZ0RCPFF_M95IQDfmdYKC0Er656UgKUK4UKG9JlxP4o81UwoJoZYz_D1zTlltipEa5RiNvUtNU8vLKoGSY"

        unwrapped_private_key = Key.from_paserk(wpk, password="our-secret")
        token = pyseto.encode(
            unwrapped_private_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        decoded = pyseto.decode(public_key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_paserk_seal(self):

        raw_key = Key.new(version=4, purpose="local", key=b"our-secret")
        token = pyseto.encode(
            raw_key,
            b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
        )
        with open(get_path("keys/public_key_x25519.pem")) as key_file:
            sealed_key = raw_key.to_paserk(sealing_key=key_file.read())

        with open(get_path("keys/private_key_x25519.pem")) as key_file:
            unsealed_key = Key.from_paserk(sealed_key, unsealing_key=key_file.read())
        decoded = pyseto.decode(unsealed_key, token)
        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'

    def test_sample_rtd_v4_public(self):

        with open(get_path("keys/private_key_ed25519.pem")) as key_file:
            private_key = Key.new(4, "public", key_file.read())
        token = pyseto.encode(
            private_key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
            implicit_assertion=b"xyz",  # Optional
        )

        with open(get_path("keys/public_key_ed25519.pem")) as key_file:
            public_key = Key.new(4, "public", key_file.read())
        decoded = pyseto.decode(public_key, token, implicit_assertion=b"xyz")

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v4"
        assert decoded.purpose == "public"

    def test_sample_rtd_v4_local(self):

        key = Key.new(version=4, purpose="local", key=b"our-secret")
        token = pyseto.encode(
            key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
            implicit_assertion=b"xyz",  # Optional
        )

        decoded = pyseto.decode(key, token, implicit_assertion=b"xyz")

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v4"
        assert decoded.purpose == "local"

    def test_sample_rtd_v3_public(self):

        with open(get_path("keys/private_key_ecdsa_p384.pem")) as key_file:
            private_key = Key.new(3, "public", key_file.read())
        token = pyseto.encode(
            private_key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
            implicit_assertion=b"xyz",  # Optional
        )

        with open(get_path("keys/public_key_ecdsa_p384.pem")) as key_file:
            public_key = Key.new(3, "public", key_file.read())
        decoded = pyseto.decode(public_key, token, implicit_assertion=b"xyz")

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v3"
        assert decoded.purpose == "public"

    def test_sample_rtd_v3_local(self):

        key = Key.new(version=3, purpose="local", key=b"our-secret")
        token = pyseto.encode(
            key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
            implicit_assertion=b"xyz",  # Optional
        )

        decoded = pyseto.decode(key, token, implicit_assertion=b"xyz")

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v3"
        assert decoded.purpose == "local"

    def test_sample_rtd_v2_public(self):

        with open(get_path("keys/private_key_ed25519.pem")) as key_file:
            private_key = Key.new(2, "public", key_file.read())
        token = pyseto.encode(
            private_key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
        )

        with open(get_path("keys/public_key_ed25519.pem")) as key_file:
            public_key = Key.new(2, "public", key_file.read())
        decoded = pyseto.decode(public_key, token)

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v2"
        assert decoded.purpose == "public"

    def test_sample_rtd_v2_local(self):

        key = Key.new(version=2, purpose="local", key=token_bytes(32))
        token = pyseto.encode(
            key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
        )

        decoded = pyseto.decode(key, token)

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v2"
        assert decoded.purpose == "local"

    def test_sample_rtd_v1_public(self):

        with open(get_path("keys/private_key_rsa.pem")) as key_file:
            private_key = Key.new(1, "public", key_file.read())
        token = pyseto.encode(
            private_key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
        )

        with open(get_path("keys/public_key_rsa.pem")) as key_file:
            public_key = Key.new(1, "public", key_file.read())
        decoded = pyseto.decode(public_key, token)

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v1"
        assert decoded.purpose == "public"

    def test_sample_rtd_v1_local(self):

        key = Key.new(version=1, purpose="local", key=b"our-secret")
        token = pyseto.encode(
            key,
            payload=b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}',
            footer=b"This is a footer",  # Optional
        )

        decoded = pyseto.decode(key, token)

        assert decoded.payload == b'{"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"}'
        assert decoded.footer == b"This is a footer"
        assert decoded.version == "v1"
        assert decoded.purpose == "local"



import pytest

from pyseto.utils import i2osp


class TestUtils:
    """
    Tests for utils.
    """

    def test_utils_i2osp_invalid_arg(self):
        with pytest.raises(ValueError) as err:
            i2osp(270, 1)
            pytest.fail("i2osp should fail.")
        assert "integer too large" in str(err.value)

    def test_utils_i2osp_with_padding(self):
        res = i2osp(
            1773640271034215956220647394962766686220368790999773790300657174885851868252873740857850633982047596450206042428564987609136709554719966122707327315574236228,
            66,
        )
        assert (
            res
            == b"\x00\x84H\xbd\x0e;\xc5\xb7\xdf\\\x1f\xf9\x03\xd2Db\xd1\xdf7\x1b}\x80g4A}\xb0\x19\xcd-\xc6n\x1e&\xe8\xafm\xdd!\xb6\xc3 9}\xb5%\xb3\x02\x87V\xf8\x94C\xe1O\xeb\x14f\x93\xe5\xc7\x8e}#\xa4D"
        )



from typing import Union

from .utils import base64url_decode


class Token(object):
    """
    The parsed token object which is a return value of :func:`pyseto.decode <pyseto.decode>`.
    """

    def __init__(
        self,
        version: str,
        purpose: str,
        payload: Union[bytes, dict],
        footer: Union[bytes, dict] = b"",
    ):
        self._version = version
        self._purpose = purpose
        self._payload = payload
        self._footer = footer
        self._header = (version + "." + purpose + ".").encode("utf-8")

    @classmethod
    def new(cls, token: Union[bytes, str]):
        token = token if isinstance(token, str) else token.decode("utf-8")
        t = token.split(".")
        if len(t) != 3 and len(t) != 4:
            raise ValueError("token is invalid.")
        if not t[2]:
            raise ValueError("Empty payload.")
        p = base64url_decode(t[2])
        f = base64url_decode(t[3]) if len(t) == 4 else b""
        return cls(t[0], t[1], p, f)

    @property
    def version(self) -> str:
        """
        The version of the token. It will be ``"v1"``, ``"v2"``, ``"v3"`` or ``"v4"``.
        """
        return self._version

    @property
    def purpose(self) -> str:
        """
        The purpose of the token. It will be ``"local"`` or ``"public"``.
        """
        return self._purpose

    @property
    def header(self) -> bytes:
        """
        The header of the token. It will be ``"<version>.<type>."``.
        For example, ``"v1.local."``.
        """
        return self._header

    @property
    def payload(self) -> Union[bytes, dict]:
        """
        The payload of the token which is a decoded binary string. It's not Base64 encoded data.
        """
        return self._payload

    @payload.setter
    def payload(self, payload: Union[bytes, dict]):
        """
        A setter of the payload.
        """
        self._payload = payload
        return

    @property
    def footer(self) -> Union[bytes, dict]:
        """
        The footer of the token which is a decoded binary string. It's not Base64 encoded data.
        """
        return self._footer

    @footer.setter
    def footer(self, footer: Union[bytes, dict]):
        """
        A setter of the footer.
        """
        self._footer = footer
        return



import hashlib
from secrets import token_bytes
from typing import Any, Union

from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)

from ..exceptions import DecryptError, SignError, VerifyError
from ..key_sodium import SodiumKey
from ..utils import base64url_encode, pae


class V4Local(SodiumKey):
    """
    The key object for v4.local.
    """

    _VERSION = 4
    _TYPE = "local"

    def __init__(self, key: Union[str, bytes]):

        super().__init__(key)
        if len(self._key) > 64:
            raise ValueError("key length must be up to 64 bytes.")
        return

    def to_paserk_id(self) -> str:
        h = "k4.lid."
        p = self.to_paserk()
        b = hashlib.blake2b(digest_size=33)
        b.update((h + p).encode("utf-8"))
        d = b.digest()
        return h + base64url_encode(d).decode("utf-8")

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:

        if nonce:
            if len(nonce) != 32:
                raise ValueError("nonce must be 32 bytes long.")
        else:
            nonce = token_bytes(32)
        tmp = self._generate_hash(self._key, b"paseto-encryption-key" + nonce, 56)
        ek = tmp[0:32]
        n2 = tmp[32:]
        ak = self._generate_hash(self._key, b"paseto-auth-key-for-aead" + nonce, 32)

        c = self._encrypt(ek, n2, payload)
        pre_auth = pae([self.header, nonce, c, footer, implicit_assertion])
        t = self._generate_hash(ak, pre_auth, 32)
        token = self._header + base64url_encode(nonce + c + t)
        if footer:
            token += b"." + base64url_encode(footer)
        return token

    def decrypt(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:

        n = payload[0:32]
        c = payload[32 : len(payload) - 32]
        t = payload[-32:]
        tmp = self._generate_hash(self._key, b"paseto-encryption-key" + n, 56)
        ek = tmp[0:32]
        n2 = tmp[32:]
        ak = self._generate_hash(self._key, b"paseto-auth-key-for-aead" + n, 32)
        pre_auth = pae([self.header, n, c, footer, implicit_assertion])
        t2 = self._generate_hash(ak, pre_auth, 32)
        if t != t2:
            raise DecryptError("Failed to decrypt.")
        return self._decrypt(ek, n2, c)


class V4Public(SodiumKey):
    """
    The key object for v4.public.
    """

    _VERSION = 4
    _TYPE = "public"

    def __init__(self, key: Any):

        super().__init__(key)
        self._sig_size = 64

        if not isinstance(self._key, (Ed25519PublicKey, Ed25519PrivateKey)):
            raise ValueError("The key is not Ed25519 key.")

        if isinstance(self._key, Ed25519PublicKey):
            self._is_secret = False
        return

    # @classmethod
    # def from_public_bytes(cls, key: bytes):
    #     try:
    #         k = Ed25519PublicKey.from_public_bytes(key)
    #     except Exception as err:
    #         raise ValueError("Invalid bytes for the key.") from err
    #     return cls(k)

    def to_paserk_id(self) -> str:
        p = self.to_paserk()
        h = "k4.pid." if isinstance(self._key, Ed25519PublicKey) else "k4.sid."
        b = hashlib.blake2b(digest_size=33)
        b.update((h + p).encode("utf-8"))
        d = b.digest()
        return h + base64url_encode(d).decode("utf-8")

    def sign(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b"") -> bytes:

        if isinstance(self._key, Ed25519PublicKey):
            raise ValueError("A public key cannot be used for signing.")

        m2 = pae([self.header, payload, footer, implicit_assertion])
        try:
            return self._key.sign(m2)
        except Exception as err:
            raise SignError("Failed to sign.") from err

    def verify(self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""):

        if len(payload) <= self._sig_size:
            raise ValueError("Invalid payload.")

        sig = payload[-self._sig_size :]
        m = payload[: len(payload) - self._sig_size]
        k = self._key if isinstance(self._key, Ed25519PublicKey) else self._key.public_key()
        m2 = pae([self.header, m, footer, implicit_assertion])
        try:
            k.verify(sig, m2)
        except Exception as err:
            raise VerifyError("Failed to verify.") from err
        return m



from .exceptions import (
    DecryptError,
    EncryptError,
    NotSupportedError,
    PysetoError,
    SignError,
    VerifyError,
)
from .key import Key
from .key_interface import KeyInterface
from .paseto import Paseto
from .pyseto import decode, encode
from .token import Token

__version__ = "1.6.2"
__title__ = "PySETO"
__description__ = "A Python implementation of PASETO/PASERK."
__url__ = "https://pyseto.readthedocs.io"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"
__author__ = "AJITOMI Daisuke"
__email__ = "ajitomi@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2021-2022 Ajitomi Daisuke"
__all__ = [
    "encode",
    "decode",
    "Key",
    "KeyInterface",
    "Paseto",
    "PysetoError",
    "Token",
    "DecryptError",
    "EncryptError",
    "NotSupportedError",
    "SignError",
    "VerifyError",
]



import hashlib
from secrets import token_bytes
from typing import Any, Union

from Cryptodome.Cipher import ChaCha20
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    load_pem_private_key,
    load_pem_public_key,
)
from passlib.hash import argon2

from .exceptions import DecryptError, EncryptError
from .key_interface import KeyInterface
from .utils import base64url_decode, base64url_encode


class SodiumKey(KeyInterface):
    """
    Sodium (v2, v4) PASETO key.
    """

    # _VERSION = 2 or 4
    # _TYPE = "local", "public" or "secret"

    def __init__(self, key: Any):

        super().__init__(self._VERSION, self._TYPE, key)
        return

    @classmethod
    def from_paserk(
        cls,
        paserk: str,
        wrapping_key: bytes = b"",
        password: bytes = b"",
        unsealing_key: bytes = b"",
    ) -> KeyInterface:

        if wrapping_key and password:
            raise ValueError("Only one of wrapping_key or password should be specified.")

        frags = paserk.split(".")
        if frags[0] != f"k{cls._VERSION}":
            raise ValueError(f"Invalid PASERK version: {frags[0]}.")

        if wrapping_key:
            # {local, secret}-wrap
            if len(frags) != 4:
                raise ValueError("Invalid PASERK format.")
            if frags[2] != "pie":
                raise ValueError(f"Unknown wrapping algorithm: {frags[2]}.")

            h = frags[0] + "." + frags[1] + ".pie."
            if frags[1] == "local-wrap":
                return cls(cls._decode_pie(h, wrapping_key, frags[3]))
            if frags[1] == "secret-wrap":
                k = cls._decode_pie(h, wrapping_key, frags[3])[0:32]
                return cls(Ed25519PrivateKey.from_private_bytes(k))
            raise ValueError(f"Invalid PASERK type: {frags[1]}.")

        if len(frags) != 3:
            raise ValueError("Invalid PASERK format.")

        if password:
            # {local, secret}-pw
            h = frags[0] + "." + frags[1] + "."
            if frags[1] == "local-pw":
                return cls(cls._decode_pbkw(h, password, frags[2]))
            if frags[1] == "secret-pw":
                k = cls._decode_pbkw(h, password, frags[2])[0:32]
                return cls(Ed25519PrivateKey.from_private_bytes(k))
            raise ValueError(f"Invalid PASERK type: {frags[1]}.")

        if unsealing_key:
            # seal
            h = frags[0] + "." + frags[1] + "."
            if frags[1] != "seal":
                raise ValueError(f"Invalid PASERK type: {frags[1]}.")
            if unsealing_key.startswith(b"-----BEGIN PRIVATE KEY"):
                sk = load_pem_private_key(unsealing_key, password=None)
            else:
                raise ValueError("Invalid or unsupported PEM format.")
            raw_key = sk.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            )
            k = cls._decode_pke(h, raw_key, frags[2])
            return cls(k)

        k = base64url_decode(frags[2])

        # local
        if cls._TYPE == "local":
            if frags[1] == "local":
                return cls(k)
            if frags[1] == "local-wrap":
                raise ValueError(f"{frags[1]} needs wrapping_key.")
            if frags[1] == "local-pw":
                raise ValueError(f"{frags[1]} needs password.")
            if frags[1] == "seal":
                raise ValueError(f"{frags[1]} needs unsealing_key.")
            raise ValueError(f"Invalid PASERK type: {frags[1]}.")

        # public, secret
        if frags[1] == "public":
            return cls(Ed25519PublicKey.from_public_bytes(k))
        if frags[1] == "secret":
            return cls(Ed25519PrivateKey.from_private_bytes(k[0:32]))
        if frags[1] == "secret-wrap":
            raise ValueError(f"{frags[1]} needs wrapping_key.")
        if frags[1] == "secret-pw":
            raise ValueError(f"{frags[1]} needs password.")
        raise ValueError(f"Invalid PASERK type: {frags[1]}.")

    def to_paserk(
        self,
        wrapping_key: Union[bytes, str] = b"",
        password: Union[bytes, str] = b"",
        sealing_key: Union[bytes, str] = b"",
        iteration: int = 100000,
        memory_cost: int = 15 * 1024,
        time_cost: int = 2,
        parallelism: int = 1,
    ) -> str:

        if wrapping_key and password:
            raise ValueError("Only one of wrapping_key or password should be specified.")

        if wrapping_key:
            # local-wrap
            bkey = wrapping_key if isinstance(wrapping_key, bytes) else wrapping_key.encode("utf-8")
            if self.purpose == "local":
                h = f"k{self.version}.local-wrap.pie."
                return h + self._encode_pie(h, bkey, self._key)

            # secret-wrap
            if not isinstance(self._key, Ed25519PrivateKey):
                raise ValueError("Public key cannot be wrapped.")

            h = f"k{self.version}.secret-wrap.pie."
            priv = self._key.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            )
            pub = self._key.public_key().public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            )
            return h + self._encode_pie(h, bkey, priv + pub)

        if password:
            # local-pw
            bpw = password if isinstance(password, bytes) else password.encode("utf-8")
            if self.purpose == "local":
                h = f"k{self.version}.local-pw."
                return h + self._encode_pbkw(h, bpw, self._key, memory_cost, time_cost, parallelism)

            # secret-pw
            if not isinstance(self._key, Ed25519PrivateKey):
                raise ValueError("Public key cannot be wrapped.")

            h = f"k{self.version}.secret-pw."
            priv = self._key.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            )
            pub = self._key.public_key().public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            )
            return h + self._encode_pbkw(h, bpw, priv + pub, memory_cost, time_cost, parallelism)

        if sealing_key:
            # seal
            bkey = sealing_key if isinstance(sealing_key, bytes) else sealing_key.encode("utf-8")
            if self.purpose != "local":
                raise ValueError("Key sealing can only be used for local key.")
            if bkey.startswith(b"-----BEGIN PUBLIC KEY"):
                pk = load_pem_public_key(bkey)
            else:
                raise ValueError("Invalid or unsupported PEM format.")
            raw_key = pk.public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            )
            h = f"k{self.version}.seal."
            return h + self._encode_pke(h, raw_key, self._key)

        # local
        if self.purpose == "local":
            h = f"k{self.version}.local."
            return h + base64url_encode(self._key).decode("utf-8")

        # public
        if isinstance(self._key, Ed25519PublicKey):
            h = f"k{self.version}.public."
            pub = self._key.public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            )
            return h + base64url_encode(pub).decode("utf-8")

        # secret
        h = f"k{self.version}.secret."
        priv = self._key.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
        )
        pub = self._key.public_key().public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        )
        return h + base64url_encode(priv + pub).decode("utf-8")

    def to_peer_paserk_id(self) -> str:
        if not self._is_secret:
            return ""

        h1 = f"k{self.version}.public."
        pub = self._key.public_key().public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        )
        p = h1 + base64url_encode(pub).decode("utf-8")

        h2 = f"k{self.version}.pid."
        b = hashlib.blake2b(digest_size=33)
        b.update((h2 + p).encode("utf-8"))
        d = b.digest()
        return h2 + base64url_encode(d).decode("utf-8")

    @classmethod
    def _encode_pie(cls, header: str, wrapping_key: bytes, ptk: bytes) -> str:

        h = header.encode("utf-8")
        n = token_bytes(32)
        x = cls._generate_hash(wrapping_key, b"\x80" + n, 56)
        ek = x[0:32]
        n2 = x[32:]
        ak = cls._generate_hash(wrapping_key, b"\x81" + n, 32)
        c = cls._encrypt(ek, n2, ptk)
        t = cls._generate_hash(ak, h + n + c, 32)
        return base64url_encode(t + n + c).decode("utf-8")

    @classmethod
    def _decode_pie(cls, header: str, wrapping_key: bytes, data: str) -> bytes:

        h = header.encode("utf-8")
        d = base64url_decode(data)
        t = d[0:32]
        n = d[32:64]
        c = d[64:]
        ak = cls._generate_hash(wrapping_key, b"\x81" + n, 32)

        t2 = cls._generate_hash(ak, h + n + c, 32)
        if t != t2:
            raise DecryptError("Failed to unwrap a key.")

        x = cls._generate_hash(wrapping_key, b"\x80" + n, 56)
        ek = x[0:32]
        n2 = x[32:]
        return cls._decrypt(ek, n2, c)

    @classmethod
    def _encode_pbkw(
        cls,
        header: str,
        password: bytes,
        ptk: bytes,
        memory_cost: int,
        time_cost: int,
        parallelism: int,
    ) -> str:

        h = header.encode("utf-8")
        s = token_bytes(16)
        argon2_k = argon2.using(
            salt=s,
            memory_cost=memory_cost,
            time_cost=time_cost,
            parallelism=parallelism,
            digest_size=32,
        ).hash(password)
        frags = argon2_k.split("$")
        key = base64url_decode(frags[5])
        ek = cls._digest(b"\xff" + key, 32)
        ak = cls._digest(b"\xfe" + key, 32)
        n = token_bytes(24)
        edk = cls._encrypt(ek, n, ptk)
        mem = (memory_cost * 1024).to_bytes(8, byteorder="big")
        time = time_cost.to_bytes(4, byteorder="big")
        para = parallelism.to_bytes(4, byteorder="big")
        t = cls._generate_hash(ak, h + s + mem + time + para + n + edk, 32)
        return base64url_encode(s + mem + time + para + n + edk + t).decode("utf-8")

    @classmethod
    def _decode_pbkw(cls, header: str, password: bytes, data: str) -> bytes:

        h = header.encode("utf-8")
        d = base64url_decode(data)

        s = d[0:16]
        mem = d[16:24]
        time = d[24:28]
        para = d[28:32]
        n = d[32:56]
        edk = d[56 : len(d) - 32]
        t = d[-32:]
        memory_cost = int.from_bytes(mem, byteorder="big")
        time_cost = int.from_bytes(time, byteorder="big")
        parallelism = int.from_bytes(para, byteorder="big")
        argon2_k = argon2.using(
            salt=s,
            memory_cost=int(memory_cost / 1024),
            time_cost=time_cost,
            parallelism=parallelism,
            digest_size=32,
        ).hash(password)
        frags = argon2_k.split("$")
        k = base64url_decode(frags[5])
        ak = cls._digest(b"\xfe" + k, 32)
        t2 = cls._generate_hash(ak, h + s + mem + time + para + n + edk, 32)
        if t != t2:
            raise DecryptError("Failed to unwrap a key.")

        ek = cls._digest(b"\xff" + k, 32)
        return cls._decrypt(ek, n, edk)

    @classmethod
    def _encode_pke(
        cls,
        header: str,
        sealing_key: bytes,
        ptk: bytes,
    ) -> str:

        h = header.encode("utf-8")

        xpk = X25519PublicKey.from_public_bytes(sealing_key)
        esk = X25519PrivateKey.generate()
        epk = esk.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw)
        xk = esk.exchange(xpk)

        ek = cls._digest(b"\x01" + h + xk + epk + sealing_key, 32)
        ak = cls._digest(b"\x02" + h + xk + epk + sealing_key, 32)
        n = cls._digest(epk + sealing_key, 24)

        edk = cls._encrypt(ek, n, ptk)
        t = cls._generate_hash(ak, h + epk + edk, 32)
        return base64url_encode(t + epk + edk).decode("utf-8")

    @classmethod
    def _decode_pke(cls, header: str, unsealing_key: bytes, data: str) -> bytes:

        h = header.encode("utf-8")
        d = base64url_decode(data)

        t = d[0:32]
        epk = d[32:64]
        edk = d[64:]

        xsk = X25519PrivateKey.from_private_bytes(unsealing_key)
        xpk = xsk.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw)
        xk = xsk.exchange(X25519PublicKey.from_public_bytes(epk))

        ak = cls._digest(b"\x02" + h + xk + epk + xpk, 32)
        t2 = cls._generate_hash(ak, h + epk + edk, 32)
        if t != t2:
            raise DecryptError("Failed to unseal a key.")

        ek = cls._digest(b"\x01" + h + xk + epk + xpk, 32)
        n = cls._digest(epk + xpk, 24)
        return cls._decrypt(ek, n, edk)

    @staticmethod
    def _generate_hash(key: bytes, msg: bytes, size: int) -> bytes:

        try:
            h = hashlib.blake2b(key=key, digest_size=size)
            h.update(msg)
            return h.digest()
        except Exception as err:
            raise EncryptError("Failed to generate hash.") from err

    @staticmethod
    def _digest(msg: bytes, size: int) -> bytes:

        h = hashlib.blake2b(digest_size=size)
        h.update(msg)
        return h.digest()

    @staticmethod
    def _encrypt(key: bytes, nonce: bytes, msg: bytes) -> bytes:

        try:
            cipher = ChaCha20.new(key=key, nonce=nonce)
            return cipher.encrypt(msg)
        except Exception as err:
            raise EncryptError("Failed to encrypt.") from err

    @staticmethod
    def _decrypt(key: bytes, nonce: bytes, msg: bytes) -> bytes:

        try:
            cipher = ChaCha20.new(key=key, nonce=nonce)
            return cipher.decrypt(msg)
        except Exception as err:
            raise DecryptError("Failed to decrypt.") from err



import json
import time
from datetime import datetime, timedelta, timezone

import freezegun
import pytest

import pyseto
from pyseto import Key
from pyseto.exceptions import VerifyError

from .utils import load_key


class InvalidSerializer:
    def __init__(self):
        self.dumps = "not a function."


class InvalidSerializer2:
    def dumps(self, *args):
        raise NotImplementedError("Not implemented")


class InvalidDeserializer:
    def __init__(self):
        self.loads = "not a function."


class InvalidDeserializer2:
    def loads(self, *args):
        raise NotImplementedError("Not implemented")


class TestPyseto:
    """
    Tests for pyseto.encode and decode.
    """

    @pytest.mark.parametrize(
        "version, key, msg",
        [
            (
                1,
                load_key("keys/public_key_rsa.pem"),
                "A public key cannot be used for signing.",
            ),
            (
                2,
                load_key("keys/public_key_ed25519.pem"),
                "A public key cannot be used for signing.",
            ),
            (
                3,
                load_key("keys/public_key_ecdsa_p384.pem"),
                "A public key cannot be used for signing.",
            ),
            (
                4,
                load_key("keys/public_key_ed25519.pem"),
                "A public key cannot be used for signing.",
            ),
        ],
    )
    def test_encode_with_public_key(self, version, key, msg):
        k = Key.new(version, "public", key)
        with pytest.raises(ValueError) as err:
            pyseto.encode(k, b"Hello world!")
            pytest.fail("pyseto.encode() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "serializer, msg",
        [
            (
                None,
                "serializer should be specified for the payload object.",
            ),
            (
                {},
                "serializer should be specified for the payload object.",
            ),
            (
                [],
                "serializer should be specified for the payload object.",
            ),
            (
                "",
                "serializer should be specified for the payload object.",
            ),
            (
                b"",
                "serializer should be specified for the payload object.",
            ),
            (
                {"key": "value"},
                "serializer should have dumps().",
            ),
            (
                InvalidSerializer(),
                "serializer should have dumps().",
            ),
            (
                InvalidSerializer2(),
                "Failed to serialize the payload.",
            ),
        ],
    )
    def test_encode_object_payload_with_invalid_serializer(self, serializer, msg):
        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        with pytest.raises(ValueError) as err:
            pyseto.encode(
                private_key,
                {
                    "data": "this is a signed message",
                    "exp": "2022-01-01T00:00:00+00:00",
                },
                serializer=serializer,
            )
            pytest.fail("pyseto.encode() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "serializer, msg",
        [
            (
                None,
                "serializer should be specified for the footer object.",
            ),
            (
                {},
                "serializer should be specified for the footer object.",
            ),
            (
                [],
                "serializer should be specified for the footer object.",
            ),
            (
                "",
                "serializer should be specified for the footer object.",
            ),
            (
                b"",
                "serializer should be specified for the footer object.",
            ),
            (
                {"key": "value"},
                "serializer should have dumps().",
            ),
            (
                InvalidSerializer(),
                "serializer should have dumps().",
            ),
            (
                InvalidSerializer2(),
                "Failed to serialize the footer.",
            ),
        ],
    )
    def test_encode_object_footer_with_invalid_serializer(self, serializer, msg):
        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        with pytest.raises(ValueError) as err:
            pyseto.encode(
                private_key,
                b"Hello world!",
                footer={
                    "kid": "xxxxxx",
                },
                serializer=serializer,
            )
            pytest.fail("pyseto.encode() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "version, key, msg",
        [
            (1, load_key("keys/public_key_rsa.pem"), "Invalid payload."),
            (2, load_key("keys/public_key_ed25519.pem"), "Invalid payload."),
            (3, load_key("keys/public_key_ecdsa_p384.pem"), "Invalid payload."),
            (4, load_key("keys/public_key_ed25519.pem"), "Invalid payload."),
        ],
    )
    def test_decode_with_invalid_payload(self, version, key, msg):
        k = Key.new(version, "public", key)
        with pytest.raises(ValueError) as err:
            pyseto.decode(k, f"v{version}.public.11111111")
            pytest.fail("pyseto.decode() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "version, public_key",
        [
            (1, load_key("keys/public_key_rsa.pem")),
            (2, load_key("keys/public_key_ed25519.pem")),
            (3, load_key("keys/public_key_ecdsa_p384.pem")),
        ],
    )
    def test_decode_with_another_version_key(self, version, public_key):
        sk = Key.new(4, "public", load_key("keys/private_key_ed25519.pem"))
        pk = Key.new(version, "public", public_key)
        token = pyseto.encode(sk, "Hello world!")
        with pytest.raises(ValueError) as err:
            pyseto.decode(pk, token)
            pytest.fail("pyseto.decode() should fail.")
        assert "key is not found for verifying the token." in str(err.value)

    @pytest.mark.parametrize(
        "deserializer, msg",
        [
            (
                {"key": "value"},
                "deserializer should have loads().",
            ),
            (
                InvalidDeserializer(),
                "deserializer should have loads().",
            ),
            (
                InvalidDeserializer2(),
                "Failed to deserialize the payload.",
            ),
        ],
    )
    def test_decode_object_payload_with_invalid_deserializer(self, deserializer, msg):
        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"
        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(ValueError) as err:
            pyseto.decode(public_key, token, deserializer=deserializer)
            pytest.fail("pyseto.decode() should fail.")
        assert msg in str(err.value)

    @freezegun.freeze_time("2021-01-01")
    def test_decode_bytes_footer_with_deserializer(self):
        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"
        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "exp": "2022-01-01T00:00:00+00:00"},
            footer=b"This is a footer.",
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token, deserializer=json)
        assert isinstance(decoded.footer, bytes)
        assert decoded.footer == b"This is a footer."

    def test_decode_object_payload_with_invalid_exp(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "exp": "xxxxx"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(VerifyError) as err:
            pyseto.decode(public_key, token, deserializer=json)
            pytest.fail("pyseto.decode() should fail.")
        assert "Invalid exp." in str(err.value)

    def test_decode_object_payload_with_expired_exp(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message"},
            exp=1,
        )
        time.sleep(2)
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(VerifyError) as err:
            pyseto.decode(public_key, token, deserializer=json)
            pytest.fail("pyseto.decode() should fail.")
        assert "Token expired." in str(err.value)

    def test_decode_object_payload_with_aud(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "aud": "12345"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        decoded = pyseto.decode(public_key, token, deserializer=json, aud="12345")
        assert decoded.payload["aud"] == "12345"

    def test_decode_object_payload_without_aud(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(VerifyError) as err:
            pyseto.decode(public_key, token, deserializer=json, aud="12345")
            pytest.fail("pyseto.decode() should fail.")
        assert "aud verification failed." in str(err.value)

    def test_decode_object_payload_with_invalid_aud(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "aud": "12345"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(VerifyError) as err:
            pyseto.decode(public_key, token, deserializer=json, aud="1234x")
            pytest.fail("pyseto.decode() should fail.")
        assert "aud verification failed." in str(err.value)

    def test_decode_object_payload_with_invalid_nbf(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {"data": "this is a signed message", "nbf": "xxxxx"},
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(VerifyError) as err:
            pyseto.decode(public_key, token, deserializer=json)
            pytest.fail("pyseto.decode() should fail.")
        assert "Invalid nbf." in str(err.value)

    def test_decode_object_payload_with_future_nbf(self):

        private_key_pem = b"-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----"
        public_key_pem = b"-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----"

        now = datetime.now(tz=timezone.utc)
        private_key = Key.new(version=4, purpose="public", key=private_key_pem)
        token = pyseto.encode(
            private_key,
            {
                "data": "this is a signed message",
                "nbf": (now + timedelta(seconds=10)).isoformat(timespec="seconds"),
            },
        )
        public_key = Key.new(version=4, purpose="public", key=public_key_pem)
        with pytest.raises(VerifyError) as err:
            pyseto.decode(public_key, token, deserializer=json)
            pytest.fail("pyseto.decode() should fail.")
        assert "Token has not been activated yet." in str(err.value)

    def test_decode_with_empty_list_of_keys(self):
        sk = Key.new(4, "public", load_key("keys/private_key_ed25519.pem"))
        token = pyseto.encode(sk, "Hello world!")
        with pytest.raises(ValueError) as err:
            pyseto.decode([], token)
            pytest.fail("pyseto.decode() should fail.")
        assert "key is not found for verifying the token." in str(err.value)

    def test_decode_with_different_keys(self):
        sk = Key.new(4, "public", load_key("keys/private_key_ed25519.pem"))
        pk1 = Key.new(1, "public", load_key("keys/public_key_rsa.pem"))
        pk2 = Key.new(2, "public", load_key("keys/public_key_ed25519.pem"))
        pk3 = Key.new(3, "public", load_key("keys/public_key_ecdsa_p384.pem"))
        token = pyseto.encode(sk, "Hello world!")
        with pytest.raises(ValueError) as err:
            pyseto.decode([pk1, pk2, pk3], token)
            pytest.fail("pyseto.decode() should fail.")
        assert "key is not found for verifying the token." in str(err.value)

    def test_decode_with_multiple_keys(self):
        sk = Key.new(4, "public", load_key("keys/private_key_ed25519.pem"))
        token = pyseto.encode(sk, b"Hello world!")
        pk1 = Key.new(1, "public", load_key("keys/public_key_rsa.pem"))
        pk2 = Key.new(2, "public", load_key("keys/public_key_ed25519.pem"))
        pk3 = Key.new(3, "public", load_key("keys/public_key_ecdsa_p384.pem"))
        pk4 = Key.new(4, "public", load_key("keys/public_key_ed25519.pem"))
        decoded = pyseto.decode([pk1, pk2, pk3, pk4], token)
        assert decoded.payload == b"Hello world!"

    def test_decode_with_multiple_keys_have_same_header(self):
        sk = Key.new(4, "public", load_key("keys/private_key_ed25519.pem"))
        token = pyseto.encode(sk, b"Hello world!")
        pk2 = Key.new(4, "public", load_key("keys/public_key_ed25519_2.pem"))
        pk1 = Key.new(4, "public", load_key("keys/public_key_ed25519.pem"))
        decoded = pyseto.decode([pk2, pk1], token)
        assert decoded.payload == b"Hello world!"



import json
from typing import Any, List, Optional, Union

from .key_interface import KeyInterface
from .paseto import Paseto
from .token import Token

# export
_paseto = Paseto()


def encode(
    key: KeyInterface,
    payload: Union[bytes, str, dict],
    footer: Union[bytes, str] = b"",
    implicit_assertion: Union[bytes, str] = b"",
    nonce: bytes = b"",
    serializer: Any = json,
    exp: int = 0,
) -> bytes:

    """
    Encodes a message to a PASETO token with a key for encryption or signing.

    Args:
        key (KeyInterface): A key for encryption or signing.
        payload (Union[bytes, str, dict]): A message to be encrypted or signed.
        footer (Union[bytes, str]): A footer.
        implicit_assertion (Union[bytes, str]): An implicit assertion. It is
            only used in ``v3`` or ``v4``.
        nonce (bytes): A nonce. If omitted(it's recommended), a nonce will be
            generated with ``secrets.token_bytes()`` internally. If you don't
            want ot use ``secrets.token_bytes()``, you can specify it via this
            parameter explicitly.
        serializer (Any): A serializer which is used when the type of
            ``payload`` is ``object``. It must have a ``dumps()`` function to
            serialize the payload. Typically, you can use ``json`` or ``cbor2``.
        exp (int): An expiration time (seconds) of the PASETO token. It will be
            set in the payload as the registered ``exp`` claim when serializer
            is ``json`` and this value > ``0``. If the value <= ``0``, the
            ``exp`` claim will not be set.
    Returns:
        bytes: A PASETO token.
    Raise:
        ValueError: Invalid arguments.
        EncryptError: Failed to encrypt the message.
        SignError: Failed to sign the message.
    """
    return _paseto.encode(key, payload, footer, implicit_assertion, nonce, serializer, exp)


def decode(
    keys: Union[KeyInterface, List[KeyInterface]],
    token: Union[bytes, str],
    implicit_assertion: Union[bytes, str] = b"",
    deserializer: Optional[Any] = None,
    aud: str = "",
) -> Token:

    """
    Decodes a PASETO token with a key for decryption and/or verifying.

    Args:
        keys (KeyInterface): A key for decryption or verifying the signature in the token.
        token (Union[bytes, str]): A PASETO token to be decrypted or verified.
        implicit_assertion (Union[bytes, str]): An implicit assertion. It is
            only used in ``v3`` or ``v4``.
        deserializer (Optional[Any]): A deserializer which is used when you want to
            deserialize a ``payload`` attribute in the response object. It must have a
            ``loads()`` function to deserialize the payload. Typically, you can use
            ``json`` or ``cbor2``.
        aud (str): An audience claim value for the token verification.
            If ``deserializer=json`` and the payload of the token does not
            include an ``aud`` value that matches this value, the
            verification will fail.
    Returns:
        Token: A parsed PASETO token object.
    Raise:
        ValueError: Invalid arguments.
        DecryptError: Failed to decrypt the message.
        VerifyError: Failed to verify the message.
    """
    return _paseto.decode(keys, token, implicit_assertion, deserializer, aud)



from secrets import token_bytes

import pytest

import pyseto
from pyseto import DecryptError, EncryptError, Key, VerifyError
from pyseto.versions.v2 import V2Local, V2Public

from .utils import get_path, load_key


class TestV2Local:
    """
    Tests for v2.local.
    """

    @pytest.mark.parametrize(
        "key, msg",
        [
            (b"", "key must be specified."),
            (token_bytes(1), "key must be 32 bytes long."),
            (token_bytes(8), "key must be 32 bytes long."),
            (token_bytes(16), "key must be 32 bytes long."),
            (token_bytes(31), "key must be 32 bytes long."),
            (token_bytes(33), "key must be 32 bytes long."),
        ],
    )
    def test_v2_local_new_with_invalid_arg(self, key, msg):
        with pytest.raises(ValueError) as err:
            Key.new(2, "local", key)
            pytest.fail("Key.new() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "key",
        [
            None,
            0,
            token_bytes(65),
        ],
    )
    def test_v2_local__generate_hash_with_invalid_arg(self, key):
        with pytest.raises(EncryptError) as err:
            V2Local._generate_hash(key, b"Hello world!", 32)
            pytest.fail("V2Local._generate_hash() should fail.")
        assert "Failed to generate hash." in str(err.value)

    @pytest.mark.parametrize(
        "ptk",
        [
            None,
            0,
        ],
    )
    def test_v2_local__encode_pie_with_invalid_ptk(self, ptk):
        with pytest.raises(EncryptError) as err:
            V2Local._encode_pie("v2.local-wrap.pie.", token_bytes(32), ptk)
            pytest.fail("V2Local._encode_pie() should fail.")
        assert "Failed to encrypt." in str(err.value)

    def test_v2_local_decrypt_via_decode_with_wrong_key(self):
        k1 = Key.new(2, "local", token_bytes(32))
        k2 = Key.new(2, "local", token_bytes(32))
        token = pyseto.encode(k1, b"Hello world!")
        with pytest.raises(DecryptError) as err:
            pyseto.decode(k2, token)
            pytest.fail("pyseto.decode() should fail.")
        assert "Failed to decrypt." in str(err.value)

    def test_v2_local_encrypt_with_invalid_arg(self):
        k = Key.new(2, "local", token_bytes(32))
        with pytest.raises(EncryptError) as err:
            k.encrypt(None)
            pytest.fail("pyseto.encrypt() should fail.")
        assert "Failed to generate internal nonce." in str(err.value)

    @pytest.mark.parametrize(
        "nonce",
        [
            token_bytes(1),
            token_bytes(8),
            token_bytes(23),
            token_bytes(25),
            token_bytes(32),
        ],
    )
    def test_v2_local_encrypt_via_encode_with_wrong_nonce(self, nonce):
        k = Key.new(2, "local", token_bytes(32))
        with pytest.raises(ValueError) as err:
            pyseto.encode(k, b"Hello world!", nonce=nonce)
            pytest.fail("pyseto.encode() should fail.")
        assert "nonce must be 24 bytes long." in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.local.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k3.local.AAAAAAAAAAAAAAAA", "Invalid PASERK version: k3."),
            ("k2.local.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k2.public.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k2.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
            ("k2.public.AAAAAAAAAAAAAAAA", "Invalid PASERK type: public."),
            (
                "k2.local-wrap.AAAAAAAAAAAAAAAA",
                "local-wrap needs wrapping_key.",
            ),
            (
                "k2.secret-wrap.AAAAAAAAAAAAAAAA",
                "Invalid PASERK type: secret-wrap.",
            ),
            (
                "k2.local-pw.AAAAAAAAAAAAAAAA",
                "local-pw needs password.",
            ),
            (
                "k2.seal.AAAAAAAAAAAAAAAA",
                "seal needs unsealing_key.",
            ),
        ],
    )
    def test_v2_local_from_paserk_with_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V2Local.from_paserk(paserk)
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.local-wrap.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k2.local-wrap.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k2.local-wrap.xxx.AAAAAAAAAAAAAAAA", "Unknown wrapping algorithm: xxx."),
            ("k2.xxx.pie.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
        ],
    )
    def test_v2_local_from_paserk_with_wrapping_key_and_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V2Local.from_paserk(paserk, wrapping_key=token_bytes(32))
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("k2.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
            ("k2.seal.AAAAAAAAAAAAAAAA", "Invalid or unsupported PEM format."),
        ],
    )
    def test_v2_local_from_paserk_with_unsealing_key_and_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V2Local.from_paserk(paserk, unsealing_key=token_bytes(32))
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)

    def test_v2_local_to_paserk_with_invalid_sealing_key(self):

        k = Key.new(2, "local", token_bytes(32))
        with pytest.raises(ValueError) as err:
            k.to_paserk(sealing_key=b"not-PEM-formatted-key")
            pytest.fail("Key.from_paserk should fail.")
        assert "Invalid or unsupported PEM format." in str(err.value)

    def test_v2_local_from_paserk_with_wrong_unsealing_key(self):

        k = Key.new(2, "local", token_bytes(32))
        with open(get_path("keys/public_key_x25519.pem")) as key_file:
            sealed_key = k.to_paserk(sealing_key=key_file.read())

        with open(get_path("keys/private_key_x25519_2.pem")) as key_file:
            unsealing_key = key_file.read()

        with pytest.raises(DecryptError) as err:
            Key.from_paserk(sealed_key, unsealing_key=unsealing_key)
            pytest.fail("Key.from_paserk should fail.")
        assert "Failed to unseal a key." in str(err.value)


class TestV2Public:
    """
    Tests for v2.public.
    """

    def test_v2_public_to_paserk_id(self):
        sk = Key.new(2, "public", load_key("keys/private_key_ed25519.pem"))
        pk = Key.new(2, "public", load_key("keys/public_key_ed25519.pem"))
        assert sk.to_peer_paserk_id() == pk.to_paserk_id()
        assert pk.to_peer_paserk_id() == ""

    def test_v2_public_verify_via_encode_with_wrong_key(self):
        sk = Key.new(2, "public", load_key("keys/private_key_ed25519.pem"))
        pk = Key.new(2, "public", load_key("keys/public_key_ed25519_2.pem"))
        token = pyseto.encode(sk, b"Hello world!")
        with pytest.raises(VerifyError) as err:
            pyseto.decode(pk, token)
            pytest.fail("pyseto.decode() should fail.")
        assert "Failed to verify." in str(err.value)

    def test_v2_public_to_paserk_with_sealing_key(self):
        k = Key.new(2, "public", load_key("keys/private_key_ed25519.pem"))
        with pytest.raises(ValueError) as err:
            k.to_paserk(sealing_key=b"xxx")
            pytest.fail("pyseto.to_paserk() should fail.")
        assert "Key sealing can only be used for local key." in str(err.value)

    # def test_v2_public_from_paserk_with_wrong_unsealing_key(self):
    #     key = Key.new(2, "local", token_bytes(32))
    #     pk = Key.new(2, "public", load_key("keys/public_key_ed25519.pem"))
    #     sealing_key = pk.public_bytes(Encoding.Raw, PublicFormat.Raw)
    #     sealed = key.to_paserk(sealing_key=sealing_key)
    #     sk = Key.new(2, "public", load_key("keys/private_key_ed25519_2.pem"))
    #     with pytest.raises(ValueError) as err:
    #         Key.from_paserk(unsealing_key=unsealing_key)
    #         pytest.fail("pyseto.from_paserk() should fail.")
    #     assert "Failed to unseal a key." in str(err.value)

    @pytest.mark.parametrize(
        "paserk, msg",
        [
            ("xx.public.AAAAAAAAAAAAAAAA", "Invalid PASERK version: xx."),
            ("k3.public.AAAAAAAAAAAAAAAA", "Invalid PASERK version: k3."),
            ("k2.public.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k2.local.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK format."),
            ("k2.xxx.AAAAAAAAAAAAAAAA", "Invalid PASERK type: xxx."),
            ("k2.local.AAAAAAAAAAAAAAAA", "Invalid PASERK type: local."),
            (
                "k2.local-wrap.AAAAAAAAAAAAAAAA",
                "Invalid PASERK type: local-wrap.",
            ),
            (
                "k2.secret-wrap.AAAAAAAAAAAAAAAA",
                "secret-wrap needs wrapping_key.",
            ),
            (
                "k2.secret-pw.AAAAAAAAAAAAAAAA",
                "secret-pw needs password.",
            ),
        ],
    )
    def test_v2_public_from_paserk_with_invalid_args(self, paserk, msg):

        with pytest.raises(ValueError) as err:
            V2Public.from_paserk(paserk)
            pytest.fail("Key.from_paserk should fail.")
        assert msg in str(err.value)
